from pathlib import Path
from typing import Dict, List, Optional, Union

import metatensor.torch
import torch
from metatensor.torch import Labels, TensorBlock, TensorMap
from metatensor.torch.atomistic import (
    MetatensorAtomisticModel,
    ModelCapabilities,
    ModelMetadata,
    ModelOutput,
    NeighborListOptions,
    System,
)

from metatrain.utils.data import DatasetInfo
from metatrain.utils.data.target_info import is_auxiliary_output

from ..utils.additive import ZBL
from ..utils.dtype import dtype_to_str
from ..utils.metadata import append_metadata_references
from ..utils.sum_over_atoms import sum_over_atoms
from .modules.pet import PET as RawPET
from .utils import load_raw_pet_model, systems_to_batch_dict


class PET(torch.nn.Module):
    __supported_devices__ = ["cuda", "cpu"]
    __supported_dtypes__ = [torch.float32]
    __default_metadata__ = ModelMetadata(
        references={"architecture": ["https://arxiv.org/abs/2305.19302v3"]}
    )

    def __init__(self, model_hypers: Dict, dataset_info: DatasetInfo) -> None:
        super().__init__()
        if len(dataset_info.targets) != 1:
            raise ValueError("PET only supports a single target")
        self.target_name = next(iter(dataset_info.targets.keys()))
        target = dataset_info.targets[self.target_name]
        if not (
            target.is_scalar
            and target.quantity == "energy"
            and len(target.layout.block(0).properties) == 1
        ):
            raise ValueError(
                "PET only supports total-energy-like outputs, "
                f"but a {target.quantity} was provided"
            )
        if target.per_atom:
            raise ValueError(
                "PET only supports per-structure outputs, "
                "but a per-atom output was provided"
            )

        model_hypers["D_OUTPUT"] = 1
        model_hypers["TARGET_TYPE"] = "atomic"
        model_hypers["TARGET_AGGREGATION"] = "sum"
        for key in ["R_CUT", "CUTOFF_DELTA", "RESIDUAL_FACTOR"]:
            model_hypers[key] = float(model_hypers[key])
        self.hypers = model_hypers
        self.cutoff = float(self.hypers["R_CUT"])
        self.atomic_types: List[int] = dataset_info.atomic_types
        self.dataset_info = dataset_info
        self.pet = None
        self.is_lora_applied = False
        self.checkpoint_path: Optional[str] = None

        # last-layer feature size (for LLPR module)
        self.last_layer_feature_size = (
            self.hypers["N_GNN_LAYERS"]
            * self.hypers["HEAD_N_NEURONS"]
            * (1 + self.hypers["USE_BOND_ENERGIES"])
        )
        # if they are enabled, the edge features are concatenated
        # to the node features

        # additive models: these are handled by the trainer at training
        # time, and they are added to the output at evaluation time
        additive_models = []
        if self.hypers["USE_ZBL"]:
            additive_models.append(
                ZBL(
                    {},
                    dataset_info=DatasetInfo(
                        length_unit=dataset_info.length_unit,
                        atomic_types=self.atomic_types,
                        targets={
                            target_name: target_info
                            for target_name, target_info in dataset_info.targets.items()
                            if ZBL.is_valid_target(target_name, target_info)
                        },
                    ),
                )
            )
        self.additive_models = torch.nn.ModuleList(additive_models)

    def restart(self, dataset_info: DatasetInfo) -> "PET":
        merged_info = self.dataset_info.union(dataset_info)
        new_atomic_types = [
            at for at in merged_info.atomic_types if at not in self.atomic_types
        ]
        new_targets = {
            key: value
            for key, value in merged_info.targets.items()
            if key not in self.dataset_info.targets
        }

        if len(new_atomic_types) > 0:
            raise ValueError(
                f"New atomic types found in the dataset: {new_atomic_types}. "
                "The PET model does not support adding new atomic types."
            )

        if len(new_targets) > 0:
            raise ValueError(
                f"New targets found in the training options: {new_targets}. "
                "The PET model does not support adding new training targets."
            )

        self.dataset_info = merged_info
        self.atomic_types = sorted(self.atomic_types)
        return self

    def set_trained_model(self, trained_model: RawPET):
        self.pet = trained_model  # type: ignore

    def requested_neighbor_lists(
        self,
    ) -> List[NeighborListOptions]:
        return [
            NeighborListOptions(
                cutoff=self.cutoff,
                full_list=True,
                strict=True,
            )
        ]

    def forward(
        self,
        systems: List[System],
        outputs: Dict[str, ModelOutput],
        selected_atoms: Optional[Labels] = None,
    ) -> Dict[str, TensorMap]:
        options = self.requested_neighbor_lists()[0]
        batch = systems_to_batch_dict(
            systems, options, self.atomic_types, selected_atoms
        )

        output = self.pet(batch)  # type: ignore
        predictions = output["prediction"]
        output_quantities: Dict[str, TensorMap] = {}

        structure_index = batch["batch"]
        _, counts = torch.unique(batch["batch"], return_counts=True)
        atom_index = torch.cat(
            [torch.arange(count, device=predictions.device) for count in counts]
        )
        samples_values = torch.stack([structure_index, atom_index], dim=1)
        samples = Labels(names=["system", "atom"], values=samples_values)
        empty_labels = Labels(
            names=["_"], values=torch.tensor([[0]], device=predictions.device)
        )

        # output the last-layer features for the outputs, if requested:
        if (
            f"mtt::aux::{self.target_name}_last_layer_features" in outputs
            or "features" in outputs
        ):
            ll_output_name = f"mtt::aux::{self.target_name}_last_layer_features"
            base_name = self.target_name
            if ll_output_name in outputs and base_name not in outputs:
                raise ValueError(
                    f"Features {ll_output_name} can only be requested "
                    f"if the corresponding output {base_name} is also requested."
                )
            ll_features = output["last_layer_features"]
            block = TensorBlock(
                values=ll_features,
                samples=samples,
                components=[],
                properties=Labels(
                    names=["properties"],
                    values=torch.arange(
                        ll_features.shape[1], device=predictions.device
                    ).reshape(-1, 1),
                ),
            )
            output_tmap = TensorMap(
                keys=empty_labels,
                blocks=[block],
            )
            if ll_output_name in outputs:
                ll_features_options = outputs[ll_output_name]
                if not ll_features_options.per_atom:
                    processed_output_tmap = sum_over_atoms(output_tmap)
                else:
                    processed_output_tmap = output_tmap
                output_quantities[ll_output_name] = processed_output_tmap
            if "features" in outputs:
                features_options = outputs["features"]
                if not features_options.per_atom:
                    processed_output_tmap = sum_over_atoms(output_tmap)
                else:
                    processed_output_tmap = output_tmap
                output_quantities["features"] = processed_output_tmap

        for output_name in outputs:
            if is_auxiliary_output(output_name):
                continue  # skip auxiliary outputs (not targets)
            energy_labels = Labels(
                names=["energy"], values=torch.tensor([[0]], device=predictions.device)
            )
            block = TensorBlock(
                samples=samples,
                components=[],
                properties=energy_labels,
                values=predictions,
            )
            if selected_atoms is not None:
                block = metatensor.torch.slice_block(block, "samples", selected_atoms)
            output_tmap = TensorMap(keys=empty_labels, blocks=[block])
            if not outputs[output_name].per_atom:
                output_tmap = sum_over_atoms(output_tmap)
            output_quantities[output_name] = output_tmap

        if not self.training:
            # at evaluation, we also add the additive contributions
            for additive_model in self.additive_models:
                outputs_for_additive_model: Dict[str, ModelOutput] = {}
                for output_name, output_options in outputs.items():
                    if output_name in additive_model.outputs:
                        outputs_for_additive_model[output_name] = output_options
                additive_contributions = additive_model(
                    systems,
                    outputs_for_additive_model,
                    selected_atoms,
                )
                for output_name in additive_contributions:
                    output_quantities[output_name] = metatensor.torch.add(
                        output_quantities[output_name],
                        additive_contributions[output_name],
                    )

        return output_quantities

    @classmethod
    def load_checkpoint(cls, path: Union[str, Path]) -> "PET":
        checkpoint = torch.load(path, weights_only=False, map_location="cpu")
        hypers = checkpoint["hypers"]
        model_hypers = hypers["ARCHITECTURAL_HYPERS"]
        dataset_info = checkpoint["dataset_info"]
        model = cls(model_hypers=model_hypers, dataset_info=dataset_info)
        state_dict = checkpoint["model_state_dict"]
        dtype = next(iter(state_dict.values())).dtype
        lora_state_dict = checkpoint["lora_state_dict"]
        if lora_state_dict is not None:
            model.is_lora_applied = True
        else:
            lora_state_dict = {}
        wrapper = load_raw_pet_model(
            state_dict,
            model.hypers,
            model.atomic_types,
            checkpoint["self_contributions"],
            use_lora_peft=model.is_lora_applied,
            **lora_state_dict,
        )

        model.to(dtype).set_trained_model(wrapper)

        return model

    def export(
        self, metadata: Optional[ModelMetadata] = None
    ) -> MetatensorAtomisticModel:
        dtype = next(self.parameters()).dtype
        if dtype not in self.__supported_dtypes__:
            raise ValueError(f"Unsupported dtype {self.dtype} for PET")

        # Make sure the model is all in the same dtype
        # For example, after training, the additive models could still be in
        # float64
        self.to(dtype)

        interaction_ranges = [self.hypers["N_GNN_LAYERS"] * self.cutoff]
        for additive_model in self.additive_models:
            if hasattr(additive_model, "cutoff_radius"):
                interaction_ranges.append(additive_model.cutoff_radius)
        interaction_range = max(interaction_ranges)

        capabilities = ModelCapabilities(
            outputs={
                self.target_name: ModelOutput(
                    quantity=self.dataset_info.targets[self.target_name].quantity,
                    unit=self.dataset_info.targets[self.target_name].unit,
                    per_atom=False,
                ),
                f"mtt::aux::{self.target_name.replace('mtt::', '')}_last_layer_features": ModelOutput(  # noqa: E501
                    unit="unitless", per_atom=True
                ),
            },
            atomic_types=self.atomic_types,
            interaction_range=interaction_range,
            length_unit=self.dataset_info.length_unit,
            supported_devices=["cpu", "cuda"],  # and not __supported_devices__
            dtype=dtype_to_str(dtype),
        )

        if metadata is None:
            metadata = ModelMetadata()

        append_metadata_references(metadata, self.__default_metadata__)

        return MetatensorAtomisticModel(self.eval(), metadata, capabilities)



from .model import NanoPET
from .trainer import Trainer


__model__ = NanoPET
__trainer__ = Trainer

__authors__ = [
    ("Filippo Bigi <filippo.bigi@epfl.ch>", "@frostedoyster"),
]

__maintainers__ = [
    ("Filippo Bigi <filippo.bigi@epfl.ch>", "@frostedoyster"),
]



from typing import Dict

import torch


class Encoder(torch.nn.Module):
    """
    An encoder of edges. It generates a fixed-size representation of the
    interatomic vector, the chemical element of the center and the chemical
    element of the neighbor. The representations are then concatenated and
    compressed to the initial fixed size.
    """

    def __init__(
        self,
        n_species: int,
        hidden_size: int,
    ):
        super().__init__()

        self.cartesian_encoder = torch.nn.Sequential(
            torch.nn.Linear(in_features=3, out_features=4 * hidden_size, bias=False),
            torch.nn.SiLU(),
            torch.nn.Linear(
                in_features=4 * hidden_size, out_features=4 * hidden_size, bias=False
            ),
            torch.nn.SiLU(),
            torch.nn.Linear(
                in_features=4 * hidden_size, out_features=hidden_size, bias=False
            ),
        )
        self.center_encoder = torch.nn.Embedding(
            num_embeddings=n_species, embedding_dim=hidden_size
        )
        self.neighbor_encoder = torch.nn.Embedding(
            num_embeddings=n_species, embedding_dim=hidden_size
        )
        self.compressor = torch.nn.Linear(
            in_features=3 * hidden_size, out_features=hidden_size, bias=False
        )

    def forward(
        self,
        features: Dict[str, torch.Tensor],
    ):
        # Encode cartesian coordinates
        cartesian_features = self.cartesian_encoder(features["cartesian"])

        # Encode centers
        center_features = self.center_encoder(features["center"])

        # Encode neighbors
        neighbor_features = self.neighbor_encoder(features["neighbor"])

        # Concatenate
        encoded_features = torch.concatenate(
            [cartesian_features, center_features, neighbor_features], dim=-1
        )

        # Compress
        compressed_features = self.compressor(encoded_features)

        return compressed_features



import warnings
from typing import Dict, List, Union

import torch
from metatensor.torch import Labels, TensorBlock, TensorMap
from metatensor.torch.atomistic import (
    MetatensorAtomisticModel,
    ModelEvaluationOptions,
    ModelOutput,
    System,
    is_atomistic_model,
)

from .data import TargetInfo
from .output_gradient import compute_gradient


def evaluate_model(
    model: Union[
        torch.nn.Module,
        MetatensorAtomisticModel,
        torch.jit._script.RecursiveScriptModule,
    ],
    systems: List[System],
    targets: Dict[str, TargetInfo],
    is_training: bool,
    check_consistency: bool = False,
) -> Dict[str, TensorMap]:
    """
    Evaluate the model (in training or exported) on a set of requested targets.

    :param model: The model to use. This can either be a model in training
        (``torch.nn.Module``) or an exported model
        (``torch.jit._script.RecursiveScriptModule``).
    :param systems: The systems to use.
    :param targets: The names of the targets to evaluate (keys), along with
        their associated gradients (values).
    :param is_training: Whether the model is being computed during training.

    :returns: The predictions of the model for the requested targets.
    """

    model_outputs = _get_outputs(model)
    # Assert that all targets are within the model's capabilities:
    if not set(targets.keys()).issubset(model_outputs.keys()):
        raise ValueError("Not all targets are within the model's capabilities.")

    # Find if there are any energy targets that require gradients:
    energy_targets = []
    energy_targets_that_require_position_gradients = []
    energy_targets_that_require_strain_gradients = []
    for target_name in targets.keys():
        # Check if the target is an energy:
        if model_outputs[target_name].quantity == "energy":
            energy_targets.append(target_name)
            # Check if the energy requires gradients:
            if "positions" in targets[target_name].gradients:
                energy_targets_that_require_position_gradients.append(target_name)
            if "strain" in targets[target_name].gradients:
                energy_targets_that_require_strain_gradients.append(target_name)

    new_systems = []
    strains = []
    for system in systems:
        with warnings.catch_warnings():
            # this seems to be the only way to filter out the torch-scripted warnings
            # about neighbors (which are not relevant here), regex fails
            warnings.simplefilter("ignore")
            new_system, strain = _prepare_system(
                system,
                positions_grad=len(energy_targets_that_require_position_gradients) > 0,
                strain_grad=len(energy_targets_that_require_strain_gradients) > 0,
                check_consistency=check_consistency,
            )
        new_systems.append(new_system)
        strains.append(strain)
    systems = new_systems

    # Based on the keys of the targets, get the outputs of the model:
    model_outputs = _get_model_outputs(model, systems, targets, check_consistency)

    for energy_target in energy_targets:
        # If the energy target requires gradients, compute them:
        target_requires_pos_gradients = (
            energy_target in energy_targets_that_require_position_gradients
        )
        target_requires_strain_gradients = (
            energy_target in energy_targets_that_require_strain_gradients
        )
        if target_requires_pos_gradients and target_requires_strain_gradients:
            gradients = compute_gradient(
                model_outputs[energy_target].block().values,
                [system.positions for system in systems] + strains,
                is_training=is_training,
            )
            old_energy_tensor_map = model_outputs[energy_target]
            new_block = old_energy_tensor_map.block().copy()
            new_block.add_gradient(
                "positions", _position_gradients_to_block(gradients[: len(systems)])
            )
            new_block.add_gradient(
                "strain",
                _strain_gradients_to_block(gradients[len(systems) :]),
            )
            new_energy_tensor_map = TensorMap(
                keys=old_energy_tensor_map.keys,
                blocks=[new_block],
            )
            model_outputs[energy_target] = new_energy_tensor_map
        elif target_requires_pos_gradients:
            gradients = compute_gradient(
                model_outputs[energy_target].block().values,
                [system.positions for system in systems],
                is_training=is_training,
            )
            old_energy_tensor_map = model_outputs[energy_target]
            new_block = old_energy_tensor_map.block().copy()
            new_block.add_gradient("positions", _position_gradients_to_block(gradients))
            new_energy_tensor_map = TensorMap(
                keys=old_energy_tensor_map.keys,
                blocks=[new_block],
            )
            model_outputs[energy_target] = new_energy_tensor_map
        elif target_requires_strain_gradients:
            gradients = compute_gradient(
                model_outputs[energy_target].block().values,
                strains,
                is_training=is_training,
            )
            old_energy_tensor_map = model_outputs[energy_target]
            new_block = old_energy_tensor_map.block().copy()
            new_block.add_gradient("strain", _strain_gradients_to_block(gradients))
            new_energy_tensor_map = TensorMap(
                keys=old_energy_tensor_map.keys,
                blocks=[new_block],
            )
            model_outputs[energy_target] = new_energy_tensor_map
        else:
            pass

    return model_outputs


def _position_gradients_to_block(gradients_list):
    """Convert a list of position gradients to a `TensorBlock`
    which can act as a gradient block to an energy block."""

    # `gradients` consists of a list of tensors where the second dimension is 3
    gradients = torch.concatenate(gradients_list, dim=0).unsqueeze(-1)
    # unsqueeze for the property dimension

    samples = Labels(
        names=["sample", "atom"],
        values=torch.stack(
            [
                torch.concatenate(
                    [
                        torch.tensor([i] * len(system))
                        for i, system in enumerate(gradients_list)
                    ]
                ),
                torch.concatenate(
                    [torch.arange(len(system)) for system in gradients_list]
                ),
            ],
            dim=1,
        ),
    )

    components = [
        Labels(
            names=["xyz"],
            values=torch.tensor([[0], [1], [2]]),
        )
    ]

    return TensorBlock(
        values=gradients,
        samples=samples.to(gradients.device),
        components=[c.to(gradients.device) for c in components],
        properties=Labels("energy", torch.tensor([[0]])).to(gradients.device),
    )


def _strain_gradients_to_block(gradients_list):
    """Convert a list of strain gradients to a `TensorBlock`
    which can act as a gradient block to an energy block."""

    gradients = torch.stack(gradients_list, dim=0).unsqueeze(-1)
    # unsqueeze for the property dimension

    samples = Labels(
        names=["sample"], values=torch.arange(len(gradients_list)).unsqueeze(-1)
    )

    components = [
        Labels(
            names=["xyz_1"],
            values=torch.tensor([[0], [1], [2]]),
        ),
        Labels(
            names=["xyz_2"],
            values=torch.tensor([[0], [1], [2]]),
        ),
    ]

    return TensorBlock(
        values=gradients,
        samples=samples.to(gradients.device),
        components=[c.to(gradients.device) for c in components],
        properties=Labels("energy", torch.tensor([[0]])).to(gradients.device),
    )


def _get_outputs(
    model: Union[torch.nn.Module, torch.jit._script.RecursiveScriptModule],
):
    if is_atomistic_model(model):
        return model.capabilities().outputs
    else:
        return model.outputs


def _get_model_outputs(
    model: Union[
        torch.nn.Module,
        MetatensorAtomisticModel,
        torch.jit._script.RecursiveScriptModule,
    ],
    systems: List[System],
    targets: Dict[str, TargetInfo],
    check_consistency: bool,
) -> Dict[str, TensorMap]:
    if is_atomistic_model(model):
        # put together an EvaluationOptions object
        options = ModelEvaluationOptions(
            length_unit="",  # this is only needed for unit conversions in MD engines
            outputs={
                key: ModelOutput(
                    quantity=value.quantity, unit=value.unit, per_atom=value.per_atom
                )
                for key, value in targets.items()
            },
        )
        return model(systems, options, check_consistency=check_consistency)
    else:
        return model(
            systems,
            {
                key: ModelOutput(
                    quantity=value.quantity, unit=value.unit, per_atom=value.per_atom
                )
                for key, value in targets.items()
            },
        )


@torch.jit.script
def _prepare_system(  # pragma: no cover
    system: System, positions_grad: bool, strain_grad: bool, check_consistency: bool
):
    """
    Prepares a system for gradient calculation.
    """
    if strain_grad:
        strain = torch.eye(
            3,
            dtype=system.cell.dtype,
            device=system.cell.device,
        ).requires_grad_(True)
        new_system = System(
            positions=system.positions @ strain,
            cell=system.cell @ strain,
            types=system.types,
            pbc=system.pbc,
        )
    else:
        if positions_grad:
            new_system = System(
                positions=system.positions.detach().clone().requires_grad_(True),
                cell=system.cell,
                types=system.types,
                pbc=system.pbc,
            )
            strain = None
        else:
            new_system = System(
                positions=system.positions,
                cell=system.cell,
                types=system.types,
                pbc=system.pbc,
            )
            strain = None

    for nl_options in system.known_neighbor_lists():
        nl = system.get_neighbor_list(nl_options)
        new_system.add_neighbor_list(nl_options, nl)

    return new_system, strain



import math

import numpy as np


def get_structural_batch_size(structures, atomic_batch_size):
    sizes = [len(structure.get_positions()) for structure in structures]
    average_size = np.mean(sizes)
    return math.ceil(atomic_batch_size / average_size)


def convert_atomic_throughput(train_structures, atomic_throughput):
    sizes = [len(structure.get_positions()) for structure in train_structures]
    total_size = np.sum(sizes)
    return math.ceil(atomic_throughput / total_size)


def adapt_hypers(hypers, train_structures):
    if "STRUCTURAL_BATCH_SIZE" not in hypers.__dict__.keys():
        hypers.STRUCTURAL_BATCH_SIZE = get_structural_batch_size(
            train_structures, hypers.ATOMIC_BATCH_SIZE
        )

    if "EPOCH_NUM" not in hypers.__dict__.keys():
        hypers.EPOCH_NUM = convert_atomic_throughput(
            train_structures, hypers.EPOCH_NUM_ATOMIC
        )

    if "SCHEDULER_STEP_SIZE" not in hypers.__dict__.keys():
        hypers.SCHEDULER_STEP_SIZE = convert_atomic_throughput(
            train_structures, hypers.SCHEDULER_STEP_SIZE_ATOMIC
        )

    if "EPOCHS_WARMUP" not in hypers.__dict__.keys():
        hypers.EPOCHS_WARMUP = convert_atomic_throughput(
            train_structures, hypers.EPOCHS_WARMUP_ATOMIC
        )



import torch


def dtype_to_str(dtype: torch.dtype) -> str:
    """
    Convert a torch dtype to its string representation.

    :param dtype: torch dtype to convert
    :returns: string representation of the torch dtype

    Example
    -------
    >>> import torch
    >>> dtype_to_str(torch.float64)
    "float64"
    >>> dtype_to_str(torch.int32)
    "int32"
    """
    return str(dtype).split(".")[-1]



class ArchitectureError(Exception):
    """
    Exception raised for errors originating from architectures

    This exception should be raised when an error occurs within an architecture's
    operation, indicating that the problem is not directly related to the
    metatrain infrastructure but rather to the specific architecture being used.

    :param exception: The original exception that was caught, which led to raising this
        custom exception.
    """

    def __init__(self, exception):
        super().__init__(
            f"{exception}\n\nThe error above most likely originates from an "
            "architecture.\n\nIf you think this is a bug, please contact its "
            "maintainer (see the architecture's documentation) and include the full "
            "traceback error.log."
        )



import random
from typing import Dict, List, Tuple

import numpy as np
import torch
from metatensor.torch import TensorBlock, TensorMap
from metatensor.torch.atomistic import System
from scipy.spatial.transform import Rotation

from .data import TargetInfo


def get_random_rotation():
    return Rotation.random()


def get_random_inversion():
    return random.choice([1, -1])


class RotationalAugmenter:
    def __init__(self, target_info_dict: Dict[str, TargetInfo]):
        # checks on targets
        for target_info in target_info_dict.values():
            if target_info.is_cartesian:
                if len(target_info.layout.block(0).components) != 1:
                    raise ValueError(
                        "RotationalAugmenter only supports Cartesian targets "
                        "with `rank=1`."
                    )

        self.target_info_dict = target_info_dict

        self.wigner = None
        self.complex_to_real_spherical_harmonics_transforms = {}
        is_any_target_spherical = any(
            target_info.is_spherical for target_info in target_info_dict.values()
        )
        if is_any_target_spherical:
            try:
                import spherical
            except ImportError:
                # quaternionic (used below) is a dependency of spherical
                raise ImportError(
                    "To use spherical targets with nanoPET, please install the "
                    "`spherical` package with `pip install spherical`."
                )
            largest_l = max(
                (len(block.components[0]) - 1) // 2
                for target_info in target_info_dict.values()
                if target_info.is_spherical
                for block in target_info.layout.blocks()
            )
            self.wigner = spherical.Wigner(largest_l)
            for ell in range(largest_l + 1):
                self.complex_to_real_spherical_harmonics_transforms[ell] = (
                    _complex_to_real_spherical_harmonics_transform(ell)
                )

    def apply_random_augmentations(
        self, systems: List[System], targets: Dict[str, TensorMap]
    ) -> Tuple[List[System], Dict[str, TensorMap]]:
        """
        Apply a random augmentation to a number of ``System`` objects and its targets.
        """

        rotations = [get_random_rotation() for _ in range(len(systems))]
        inversions = [get_random_inversion() for _ in range(len(systems))]
        transformations = [
            torch.from_numpy(r.as_matrix() * i) for r, i in zip(rotations, inversions)
        ]

        wigner_D_matrices = {}
        if self.wigner is not None:
            scipy_quaternions = [r.as_quat() for r in rotations]
            quaternionic_quaternions = [
                _scipy_quaternion_to_quaternionic(q) for q in scipy_quaternions
            ]
            wigner_D_matrices_complex = [
                self.wigner.D(q) for q in quaternionic_quaternions
            ]
            for target_name in targets.keys():
                target_info = self.target_info_dict[target_name]
                if target_info.is_spherical:
                    for block in target_info.layout.blocks():
                        ell = (len(block.components[0]) - 1) // 2
                        if ell not in wigner_D_matrices:  # skip if already computed
                            wigner_D_matrices_l = []
                            for wigner_D_matrix_complex in wigner_D_matrices_complex:
                                wigner_D_matrix = np.zeros(
                                    (2 * ell + 1, 2 * ell + 1), dtype=np.complex128
                                )
                                for mp in range(-ell, ell + 1):
                                    for m in range(-ell, ell + 1):
                                        wigner_D_matrix[m + ell, mp + ell] = (
                                            wigner_D_matrix_complex[
                                                self.wigner.Dindex(ell, m, mp)
                                            ]
                                        ).conj()
                                U = self.complex_to_real_spherical_harmonics_transforms[
                                    ell
                                ]
                                wigner_D_matrix = U.conj() @ wigner_D_matrix @ U.T
                                assert np.allclose(wigner_D_matrix.imag, 0.0)
                                wigner_D_matrix = wigner_D_matrix.real
                                wigner_D_matrices_l.append(
                                    torch.from_numpy(wigner_D_matrix)
                                )
                            wigner_D_matrices[ell] = wigner_D_matrices_l

        return _apply_random_augmentations(
            systems, targets, transformations, wigner_D_matrices
        )


def _apply_wigner_D_matrices(
    systems: List[System],
    target_tmap: TensorMap,
    transformations: List[torch.Tensor],
    wigner_D_matrices: Dict[int, List[torch.Tensor]],
) -> TensorMap:
    new_blocks: List[TensorBlock] = []
    for key, block in target_tmap.items():
        ell, sigma = int(key[0]), int(key[1])
        values = block.values
        if "atom" in block.samples.names:
            split_values = torch.split(
                values, [len(system.positions) for system in systems]
            )
        else:
            split_values = torch.split(values, [1 for _ in systems])
        new_values = []
        ell = (len(block.components[0]) - 1) // 2
        for v, transformation, wigner_D_matrix in zip(
            split_values, transformations, wigner_D_matrices[ell]
        ):
            is_inverted = torch.det(transformation) < 0
            new_v = v.clone()
            if is_inverted:  # inversion
                new_v = new_v * (-1) ** ell * sigma
            # fold property dimension in, apply transformation, unfold property dim
            new_v = new_v.transpose(1, 2)
            new_v = new_v @ wigner_D_matrix.T
            new_v = new_v.transpose(1, 2)
            new_values.append(new_v)
        new_values = torch.concatenate(new_values)
        new_block = TensorBlock(
            values=new_values,
            samples=block.samples,
            components=block.components,
            properties=block.properties,
        )
        new_blocks.append(new_block)

    return TensorMap(
        keys=target_tmap.keys,
        blocks=new_blocks,
    )


@torch.jit.script  # script for speed
def _apply_random_augmentations(  # pragma: no cover
    systems: List[System],
    targets: Dict[str, TensorMap],
    transformations: List[torch.Tensor],
    wigner_D_matrices: Dict[int, List[torch.Tensor]],
) -> Tuple[List[System], Dict[str, TensorMap]]:
    # Apply the transformations to the systems
    new_systems: List[System] = []
    for system, transformation in zip(systems, transformations):
        new_system = System(
            positions=system.positions @ transformation.T,
            types=system.types,
            cell=system.cell @ transformation.T,
            pbc=system.pbc,
        )
        for nl_options in system.known_neighbor_lists():
            old_nl = system.get_neighbor_list(nl_options)
            new_system.add_neighbor_list(
                nl_options,
                TensorBlock(
                    values=(old_nl.values.squeeze(-1) @ transformation.T).unsqueeze(-1),
                    samples=old_nl.samples,
                    components=old_nl.components,
                    properties=old_nl.properties,
                ),
            )
        new_systems.append(new_system)

    # Apply the transformation to the targets
    new_targets: Dict[str, TensorMap] = {}
    for name, target_tmap in targets.items():
        is_scalar = False
        if len(target_tmap.blocks()) == 1:
            if len(target_tmap.block().components) == 0:
                is_scalar = True

        is_spherical = all(
            len(block.components) == 1 and block.components[0].names == ["o3_mu"]
            for block in target_tmap.blocks()
        )

        if is_scalar:
            # no change for energies
            energy_block = TensorBlock(
                values=target_tmap.block().values,
                samples=target_tmap.block().samples,
                components=target_tmap.block().components,
                properties=target_tmap.block().properties,
            )
            if target_tmap.block().has_gradient("positions"):
                # transform position gradients:
                block = target_tmap.block().gradient("positions")
                position_gradients = block.values.squeeze(-1)
                split_sizes_forces = [system.positions.shape[0] for system in systems]
                split_position_gradients = torch.split(
                    position_gradients, split_sizes_forces
                )
                position_gradients = torch.cat(
                    [
                        split_position_gradients[i] @ transformations[i].T
                        for i in range(len(systems))
                    ]
                )
                energy_block.add_gradient(
                    "positions",
                    TensorBlock(
                        values=position_gradients.unsqueeze(-1),
                        samples=block.samples,
                        components=block.components,
                        properties=block.properties,
                    ),
                )
            if target_tmap.block().has_gradient("strain"):
                # transform strain gradients (rank 2 tensor):
                block = target_tmap.block().gradient("strain")
                strain_gradients = block.values.squeeze(-1)
                split_strain_gradients = torch.split(strain_gradients, 1)
                new_strain_gradients = torch.stack(
                    [
                        transformations[i]
                        @ split_strain_gradients[i].squeeze(0)
                        @ transformations[i].T
                        for i in range(len(systems))
                    ],
                    dim=0,
                )
                energy_block.add_gradient(
                    "strain",
                    TensorBlock(
                        values=new_strain_gradients.unsqueeze(-1),
                        samples=block.samples,
                        components=block.components,
                        properties=block.properties,
                    ),
                )
            new_targets[name] = TensorMap(
                keys=target_tmap.keys,
                blocks=[energy_block],
            )

        elif is_spherical:
            new_targets[name] = _apply_wigner_D_matrices(
                systems, target_tmap, transformations, wigner_D_matrices
            )

        else:
            # transform Cartesian vector:
            block = target_tmap.block()
            vectors = block.values
            if "atom" in target_tmap.block().samples.names:
                split_vectors = torch.split(
                    vectors, [len(system.positions) for system in systems]
                )
            else:
                split_vectors = torch.split(vectors, [1 for _ in systems])
            new_vectors = []
            for v, transformation in zip(split_vectors, transformations):
                # fold property dimension in, apply transformation, unfold property dim
                new_v = v.transpose(1, 2)
                new_v = new_v @ transformation.T
                new_v = new_v.transpose(1, 2)
                new_vectors.append(new_v)
            new_vectors = torch.cat(new_vectors)
            new_targets[name] = TensorMap(
                keys=target_tmap.keys,
                blocks=[
                    TensorBlock(
                        values=new_vectors,
                        samples=block.samples,
                        components=block.components,
                        properties=block.properties,
                    )
                ],
            )

    return new_systems, new_targets


def _complex_to_real_spherical_harmonics_transform(ell: int):
    # Generates the transformation matrix from complex spherical harmonics
    # to real spherical harmonics for a given l.
    # Returns a transformation matrix of shape ((2l+1), (2l+1)).

    if ell < 0 or not isinstance(ell, int):
        raise ValueError("l must be a non-negative integer.")

    # The size of the transformation matrix is (2l+1) x (2l+1)
    size = 2 * ell + 1
    U = np.zeros((size, size), dtype=complex)

    for m in range(-ell, ell + 1):
        m_index = m + ell  # Index in the matrix
        if m > 0:
            # Real part of Y_{l}^{m}
            U[m_index, ell + m] = 1 / np.sqrt(2) * (-1) ** m
            U[m_index, ell - m] = 1 / np.sqrt(2)
        elif m < 0:
            # Imaginary part of Y_{l}^{|m|}
            U[m_index, ell + abs(m)] = -1j / np.sqrt(2) * (-1) ** m
            U[m_index, ell - abs(m)] = 1j / np.sqrt(2)
        else:  # m == 0
            # Y_{l}^{0} remains unchanged
            U[m_index, ell] = 1

    return U


def _scipy_quaternion_to_quaternionic(q_scipy):
    # This function convert a quaternion obtained from the scipy library to the format
    # used by the quaternionic library.
    # Note: 'xyzw' is the format used by scipy.spatial.transform.Rotation
    # while 'wxyz' is the format used by quaternionic.
    qx, qy, qz, qw = q_scipy
    q_quaternion = np.array([qw, qx, qy, qz])
    return q_quaternion



from pathlib import Path

from metatrain.utils.architectures import get_default_hypers


DATASET_PATH = str(Path(__file__).parents[5] / "tests/resources/qm9_reduced_100.xyz")

DEFAULT_HYPERS = get_default_hypers("experimental.nanopet")
MODEL_HYPERS = DEFAULT_HYPERS["model"]



from typing import Dict, List, Optional, Tuple, Union

import featomic
import featomic.torch
import metatensor.torch
import numpy as np
import scipy
import torch
from metatensor import Labels, TensorBlock, TensorMap
from metatensor.torch import Labels as TorchLabels
from metatensor.torch import TensorBlock as TorchTensorBlock
from metatensor.torch import TensorMap as TorchTensorMap
from metatensor.torch.atomistic import (
    MetatensorAtomisticModel,
    ModelCapabilities,
    ModelMetadata,
    ModelOutput,
    System,
)
from skmatter._selection import _FPS as _FPS_skmatter

from metatrain.utils.data.dataset import DatasetInfo

from ..utils.additive import ZBL, CompositionModel
from ..utils.metadata import append_metadata_references


class GAP(torch.nn.Module):
    __supported_devices__ = ["cpu"]
    __supported_dtypes__ = [torch.float64]
    __default_metadata__ = ModelMetadata(
        references={
            "implementation": [
                "rascaline: https://github.com/Luthaf/rascaline",
            ],
            "architecture": [
                "SOAP: https://doi.org/10.1002/qua.24927",
                "GAP: https://doi.org/10.1103/PhysRevB.87.184115",
            ],
        }
    )

    def __init__(self, model_hypers: Dict, dataset_info: DatasetInfo) -> None:
        super().__init__()

        if len(dataset_info.targets) > 1:
            raise NotImplementedError("GAP only supports a single output")

        # Check capabilities
        for target in dataset_info.targets.values():
            if not (
                target.is_scalar
                and target.quantity == "energy"
                and len(target.layout.block(0).properties) == 1
            ):
                raise ValueError(
                    "GAP only supports total-energy-like outputs, "
                    f"but a {target.quantity} was provided"
                )
            if target.per_atom:
                raise ValueError(
                    "GAP only supports per-structure outputs, "
                    "but a per-atom output was provided"
                )
        target_name = next(iter(dataset_info.targets.keys()))
        if dataset_info.targets[target_name].quantity != "energy":
            raise ValueError("GAP only supports energies as target")
        if dataset_info.targets[target_name].per_atom:
            raise ValueError("GAP does not support per-atom energies")

        self.dataset_info = dataset_info

        self.outputs = {
            key: ModelOutput(
                quantity=value.quantity,
                unit=value.unit,
                per_atom=False,
            )
            for key, value in dataset_info.targets.items()
        }

        self.atomic_types = dataset_info.atomic_types
        self.hypers = model_hypers

        # creates a composition weight tensor that can be directly indexed by species,
        # this can be left as a tensor of zero or set from the outside using
        # set_composition_weights (recommended for better accuracy)
        n_outputs = len(dataset_info.targets)
        self.register_buffer(
            "composition_weights",
            torch.zeros(
                (n_outputs, max(self.atomic_types) + 1),
                dtype=torch.float64,  # we only support float64 for now
            ),
        )
        # buffers cannot be indexed by strings (torchscript), so we create a single
        # tensor for all output. Due to this, we need to slice the tensor when we use
        # it and use the output name to select the correct slice via a dictionary
        self.output_to_index = {
            output_name: i for i, output_name in enumerate(dataset_info.targets.keys())
        }

        self.register_buffer(
            "kernel_weights",
            torch.zeros(
                model_hypers["krr"]["num_sparse_points"],
                dtype=torch.float64,  # we only support float64 for now
            ),
        )
        # print(model_hypers["soap"])
        self._soap_torch_calculator = featomic.torch.SoapPowerSpectrum(
            **model_hypers["soap"]
        )

        kernel_kwargs = {
            "degree": model_hypers["krr"]["degree"],
            "aggregate_names": ["atom", "center_type"],
        }
        self._subset_of_regressors = SubsetOfRegressors(
            kernel_kwargs=kernel_kwargs,
        )

        self._sampler = _FPS(n_to_select=model_hypers["krr"]["num_sparse_points"])

        # set it do dummy keys, these are properly set during training
        self._keys = TorchLabels.empty("_")

        dummy_weights = TorchTensorMap(
            TorchLabels(["_"], torch.tensor([[0]])),
            [metatensor.torch.block_from_array(torch.empty(1, 1))],
        )
        dummy_X_pseudo = TorchTensorMap(
            TorchLabels(["_"], torch.tensor([[0]])),
            [metatensor.torch.block_from_array(torch.empty(1, 1))],
        )
        self._subset_of_regressors_torch = TorchSubsetofRegressors(
            dummy_weights,
            dummy_X_pseudo,
            kernel_kwargs={
                "aggregate_names": ["atom", "center_type"],
            },
        )
        self._species_labels: TorchLabels = TorchLabels.empty("_")

        # additive models: these are handled by the trainer at training
        # time, and they are added to the output at evaluation time
        composition_model = CompositionModel(
            model_hypers={},
            dataset_info=dataset_info,
        )
        additive_models = [composition_model]
        if self.hypers["zbl"]:
            additive_models.append(
                ZBL(
                    {},
                    dataset_info=DatasetInfo(
                        length_unit=dataset_info.length_unit,
                        atomic_types=self.atomic_types,
                        targets={
                            target_name: target_info
                            for target_name, target_info in dataset_info.targets.items()
                            if ZBL.is_valid_target(target_name, target_info)
                        },
                    ),
                )
            )
        self.additive_models = torch.nn.ModuleList(additive_models)

    def restart(self, dataset_info: DatasetInfo) -> "GAP":
        raise ValueError("GAP does not allow restarting training")

    def forward(
        self,
        systems: List[System],
        outputs: Dict[str, ModelOutput],
        selected_atoms: Optional[TorchLabels] = None,
    ) -> Dict[str, TorchTensorMap]:
        soap_features = self._soap_torch_calculator(
            systems, selected_samples=selected_atoms
        )
        # move keys and species labels to device
        self._keys = self._keys.to(systems[0].device)
        self._species_labels = self._species_labels.to(systems[0].device)

        new_blocks: List[TorchTensorBlock] = []
        # HACK: to add a block of zeros if there are missing species
        # which were present at training time
        # (with samples "system", "atom" = 0, 0)
        # given the values are all zeros, it does not introduce an error
        dummyblock: TorchTensorBlock = TorchTensorBlock(
            values=torch.zeros(
                (1, len(soap_features[0].properties)),
                dtype=systems[0].positions.dtype,
                device=systems[0].device,
            ),
            samples=TorchLabels(
                ["system", "atom"],
                torch.tensor([[0, 0]], dtype=torch.int, device=systems[0].device),
            ),
            properties=soap_features[0].properties,
            components=[],
        )
        if len(soap_features[0].gradients_list()) > 0:
            for idx, grad in enumerate(soap_features[0].gradients_list()):
                dummyblock_grad: TorchTensorBlock = TorchTensorBlock(
                    values=torch.zeros(
                        (
                            1,
                            soap_features[0].gradient(grad).values.shape[1 + idx],
                            len(soap_features[0].gradient(grad).properties),
                        ),
                        dtype=systems[0].positions.dtype,
                        device=systems[0].device,
                    ),
                    samples=TorchLabels(
                        ["sample", "system", "atom"],
                        torch.tensor(
                            [[0, 0, 0]], dtype=torch.int, device=systems[0].device
                        ),
                    ),
                    components=soap_features[0].gradient(grad).components,
                    properties=soap_features[0].gradient(grad).properties,
                )
                dummyblock.add_gradient(grad, dummyblock_grad)

        for idx_key in range(len(self._species_labels)):
            key = self._species_labels.entry(idx_key)
            if soap_features.keys.position(key) is not None:
                new_blocks.append(soap_features.block(key))
            else:
                new_blocks.append(dummyblock)
        soap_features = TorchTensorMap(keys=self._species_labels, blocks=new_blocks)
        soap_features = soap_features.keys_to_samples("center_type")
        # here, we move to properties to use metatensor operations to aggregate
        # later on. Perhaps we could retain the sparsity all the way to the kernels
        # of the soap features with a lot more implementation effort
        soap_features = soap_features.keys_to_properties(
            ["neighbor_1_type", "neighbor_2_type"]
        )
        soap_features = TorchTensorMap(self._keys, soap_features.blocks())
        output_key = list(outputs.keys())[0]
        energies = self._subset_of_regressors_torch(soap_features)
        return_dict = {output_key: energies}

        if not self.training:
            # at evaluation, we also add the additive contributions
            for additive_model in self.additive_models:
                outputs_for_additive_model: Dict[str, ModelOutput] = {}
                for name, output in outputs.items():
                    if name in additive_model.outputs:
                        outputs_for_additive_model[name] = output
                additive_contributions = additive_model(
                    systems,
                    outputs_for_additive_model,
                    selected_atoms,
                )
                for name in additive_contributions:
                    return_dict[name] = metatensor.torch.add(
                        return_dict[name],
                        additive_contributions[name],
                    )

        return return_dict

    def export(
        self, metadata: Optional[ModelMetadata] = None
    ) -> MetatensorAtomisticModel:
        interaction_ranges = [self.hypers["soap"]["cutoff"]["radius"]]
        for additive_model in self.additive_models:
            if hasattr(additive_model, "cutoff_radius"):
                interaction_ranges.append(additive_model.cutoff_radius)
        interaction_range = max(interaction_ranges)

        # Additionally, the composition model contains some `TensorMap`s that cannot
        # be registered correctly with Pytorch. This funciton moves them:
        self.additive_models[0]._move_weights_to_device_and_dtype(
            torch.device("cpu"), torch.float64
        )

        capabilities = ModelCapabilities(
            outputs=self.outputs,
            atomic_types=sorted(self.dataset_info.atomic_types),
            interaction_range=interaction_range,
            length_unit=self.dataset_info.length_unit,
            supported_devices=["cuda", "cpu"],
            dtype="float64",
        )

        # we export a torch scriptable regressor TorchSubsetofRegressors
        # that is used in the forward path
        self._subset_of_regressors_torch = (
            self._subset_of_regressors.export_torch_script_model()
        )

        if metadata is None:
            metadata = ModelMetadata()

        append_metadata_references(metadata, self.__default_metadata__)

        return MetatensorAtomisticModel(self.eval(), metadata, capabilities)


########################################################################################
# All classes and methods will be moved to metatensor-operations and metatensor-learn] #
########################################################################################


class _SorKernelSolver:
    """
    A few quick implementation notes, docs to be done.

    This is meant to solve the subset of regressors (SoR) problem::

    .. math::

        w = (KNM.T@KNM + reg*KMM)^-1 @ KNM.T@y

    The inverse needs to be stabilized with application of a numerical jitter,
    that is expressed as a fraction of the largest eigenvalue of KMM

    :param KMM:
        KNM matrix

    The function solve the linear problem with
    the RKHS-QR method.

    RKHS: Compute first the reproducing kernel features by diagonalizing K_MM and
          computing `P_NM = K_NM @ U_MM @ Lam_MM^(-1.2)` and then solves the linear
          problem for those (which is usually better conditioned)::

              (P_NM.T@P_NM + 1)^(-1) P_NM.T@Y
    Reference
    ---------
    Foster, L., Waagen, A., Aijaz, N., Hurley, M., Luis, A., Rinsky, J., ... &
    Srivastava, A. (2009). Stable and Efficient Gaussian Process Calculations. Journal
    of Machine Learning Research, 10(4).
    """

    def __init__(
        self,
        KMM: np.ndarray,
    ):
        self.KMM = KMM

        self._nM = len(KMM)
        self._vk, self._Uk = scipy.linalg.eigh(KMM)
        self._vk = self._vk[::-1]
        self._Uk = self._Uk[:, ::-1]

        self._nM = len(np.where(self._vk > 0)[0])
        self._PKPhi = self._Uk[:, : self._nM] * 1 / np.sqrt(self._vk[: self._nM])

    def fit(
        self, KNM: Union[torch.Tensor, np.ndarray], Y: Union[torch.Tensor, np.ndarray]
    ) -> None:
        # Convert to numpy arrays if passed as torch tensors for the solver
        if isinstance(KNM, torch.Tensor):
            weights_to_torch = True
            dtype = KNM.dtype
            device = KNM.device
            KNM = KNM.detach().numpy()
            assert isinstance(Y, torch.Tensor), "must pass `KNM` and `Y` as same type."
            Y = Y.detach().numpy()
        else:
            weights_to_torch = False

        # Broadcast Y for shape
        if len(Y.shape) == 1:
            Y = Y[:, np.newaxis]

        # Solve with the RKHS-QR method
        A = np.vstack([KNM @ self._PKPhi, np.eye(self._nM)])
        Q, R = np.linalg.qr(A)

        weights = self._PKPhi @ scipy.linalg.solve_triangular(
            R, Q.T @ np.vstack([Y, np.zeros((self._nM, Y.shape[1]))])
        )

        # Store weights as torch tensors
        if weights_to_torch:
            weights = torch.tensor(weights, dtype=dtype, device=device)
        self._weights = weights

    @property
    def weights(self):
        return self._weights

    def predict(self, KTM):
        return KTM @ self._weights


class AggregateKernel(torch.nn.Module):
    """
    A kernel that aggregates values in a kernel over :param aggregate_names: using
    the sum as aggregate function

    :param aggregate_names:

    """

    def __init__(
        self,
        aggregate_names: Union[str, List[str]],
        structurewise_aggregate: bool = False,
    ):
        super().__init__()

        self._aggregate_names = aggregate_names
        self._structurewise_aggregate = structurewise_aggregate

    def aggregate_kernel(
        self, kernel: TensorMap, are_pseudo_points: Tuple[bool, bool] = (False, False)
    ) -> TensorMap:
        if not are_pseudo_points[0]:
            kernel = metatensor.sum_over_samples(kernel, self._aggregate_names)
        if not are_pseudo_points[1]:
            raise NotImplementedError(
                "properties dimension cannot be aggregated for the moment"
            )
        return kernel

    def forward(
        self,
        tensor1: TensorMap,
        tensor2: TensorMap,
        are_pseudo_points: Tuple[bool, bool] = (False, False),
    ) -> TensorMap:
        return self.aggregate_kernel(
            self.compute_kernel(tensor1, tensor2), are_pseudo_points
        )

    def compute_kernel(self, tensor1: TensorMap, tensor2: TensorMap) -> TensorMap:
        raise NotImplementedError("compute_kernel needs to be implemented.")


class AggregatePolynomial(AggregateKernel):
    def __init__(
        self,
        aggregate_names: Union[str, List[str]],
        structurewise_aggregate: bool = False,
        degree: int = 2,
    ):
        super().__init__(aggregate_names, structurewise_aggregate)
        self._degree = degree

    def compute_kernel(self, tensor1: TensorMap, tensor2: TensorMap):
        return metatensor.pow(metatensor.dot(tensor1, tensor2), self._degree)


class TorchAggregateKernel(torch.nn.Module):
    """
    A kernel that aggregates values in a kernel over :param aggregate_names: using
    the sum as aggregate function

    :param aggregate_names:
    """

    def __init__(
        self,
        aggregate_names: Union[str, List[str]],
        structurewise_aggregate: bool = False,
    ):
        super().__init__()
        self._aggregate_names = aggregate_names
        self._structurewise_aggregate = structurewise_aggregate

    def aggregate_kernel(
        self,
        kernel: TorchTensorMap,
        are_pseudo_points: Tuple[bool, bool] = (False, False),
    ) -> TorchTensorMap:
        if not are_pseudo_points[0]:
            kernel = metatensor.torch.sum_over_samples(kernel, self._aggregate_names)
        if not are_pseudo_points[1]:
            raise NotImplementedError(
                "properties dimension cannot be aggregated for the moment"
            )
        return kernel

    def forward(
        self,
        tensor1: TorchTensorMap,
        tensor2: TorchTensorMap,
        are_pseudo_points: Tuple[bool, bool] = (False, False),
    ) -> TorchTensorMap:
        return self.aggregate_kernel(
            self.compute_kernel(tensor1, tensor2), are_pseudo_points
        )

    def compute_kernel(
        self, tensor1: TorchTensorMap, tensor2: TorchTensorMap
    ) -> TorchTensorMap:
        raise NotImplementedError("compute_kernel needs to be implemented.")


class TorchAggregatePolynomial(TorchAggregateKernel):
    def __init__(
        self,
        aggregate_names: Union[str, List[str]],
        structurewise_aggregate: bool = False,
        degree: int = 2,
    ):
        super().__init__(aggregate_names, structurewise_aggregate)
        self._degree = degree

    def compute_kernel(self, tensor1: TorchTensorMap, tensor2: TorchTensorMap):
        return metatensor.torch.pow(
            metatensor.torch.dot(tensor1, tensor2), self._degree
        )


class _FPS:
    """
    Transformer that performs Greedy Sample Selection using Farthest Point Sampling.

    Refer to :py:class:`skmatter.sample_selection.FPS` for full documentation.
    """

    def __init__(
        self,
        n_to_select=None,
    ):
        self._n_to_select = n_to_select
        self._selector_class = _FPS_skmatter
        self._selection_type = "sample"
        self._support = None

    @property
    def support(self) -> TensorMap:
        """TensorMap containing the support."""
        if self._support is None:
            raise ValueError("No selections. Call fit method first.")

        return self._support

    def fit(self, X: TensorMap):  # -> GreedySelector:
        """Learn the features to select.

        :param X:
            Training vectors.
        """
        if isinstance(X, torch.ScriptObject):
            X = torch_tensor_map_to_core(X)
            assert isinstance(X[0].values, np.ndarray)

        if len(X.component_names) != 0:
            raise ValueError("Only blocks with no components are supported.")

        blocks = []
        for _, block in X.items():
            selector = self._selector_class(
                n_to_select=self._n_to_select,
                progress_bar=False,
                score_threshold=1e-12,
                full=False,
                selection_type=self._selection_type,
            )
            selector.fit(block.values, warm_start=False)
            mask = selector.get_support()

            if self._selection_type == "feature":
                samples = Labels.single()
                properties = Labels(
                    names=block.properties.names, values=block.properties.values[mask]
                )
            elif self._selection_type == "sample":
                samples = Labels(
                    names=block.samples.names, values=block.samples.values[mask]
                )
                properties = Labels.single()

            blocks.append(
                TensorBlock(
                    values=np.zeros([len(samples), len(properties)], dtype=np.int32),
                    samples=samples,
                    components=[],
                    properties=properties,
                )
            )
        self._support = TensorMap(X.keys, blocks)

        return self

    def transform(self, X: TensorMap) -> TensorMap:
        """Reduce X to the selected features.

        :param X:
            The input tensor.
        :returns:
            The selected subset of the input.
        """
        if isinstance(X, torch.ScriptObject):
            use_mts_torch = True
            X = torch_tensor_map_to_core(X)
        else:
            use_mts_torch = False

        blocks = []
        for key, block in X.items():
            block_support = self.support.block(key)

            if self._selection_type == "feature":
                new_block = metatensor.slice_block(
                    block, "properties", block_support.properties
                )
            elif self._selection_type == "sample":
                new_block = metatensor.slice_block(
                    block, "samples", block_support.samples
                )
            blocks.append(new_block)

        X_reduced = TensorMap(X.keys, blocks)
        if use_mts_torch:
            X_reduced = core_tensor_map_to_torch(X_reduced)
        return X_reduced

    def fit_transform(self, X: TensorMap) -> TensorMap:
        """Fit to data, then transform it.

        :param X:
            Training vectors.
        """
        return self.fit(X).transform(X)


def torch_tensor_map_to_core(torch_tensor: TorchTensorMap):
    torch_blocks = []
    for _, torch_block in torch_tensor.items():
        torch_blocks.append(torch_tensor_block_to_core(torch_block))
    torch_keys = torch_labels_to_core(torch_tensor.keys)
    return TensorMap(torch_keys, torch_blocks)


def torch_tensor_block_to_core(torch_block: TorchTensorBlock):
    """Transforms a tensor block from metatensor-torch to metatensor-torch
    :param torch_block:
        tensor block from metatensor-torch
    :returns torch_block:
        tensor block from metatensor-torch
    """
    block = TensorBlock(
        values=torch_block.values.detach().cpu().numpy(),
        samples=torch_labels_to_core(torch_block.samples),
        components=[
            torch_labels_to_core(component) for component in torch_block.components
        ],
        properties=torch_labels_to_core(torch_block.properties),
    )
    for parameter, gradient in torch_block.gradients():
        block.add_gradient(
            parameter=parameter,
            gradient=TensorBlock(
                values=gradient.values.detach().cpu().numpy(),
                samples=torch_labels_to_core(gradient.samples),
                components=[
                    torch_labels_to_core(component) for component in gradient.components
                ],
                properties=torch_labels_to_core(gradient.properties),
            ),
        )
    return block


def torch_labels_to_core(torch_labels: TorchLabels):
    """Transforms labels from metatensor-torch to metatensor-torch
    :param torch_block:
        tensor block from metatensor-torch
    :returns torch_block:
        labels from metatensor-torch
    """
    return Labels(torch_labels.names, torch_labels.values.detach().cpu().numpy())


###


def core_tensor_map_to_torch(core_tensor: TensorMap):
    """Transforms a tensor map from metatensor-core to metatensor-torch
    :param core_tensor:
        tensor map from metatensor-core
    :returns torch_tensor:
        tensor map from metatensor-torch
    """

    torch_blocks = []
    for _, core_block in core_tensor.items():
        torch_blocks.append(core_tensor_block_to_torch(core_block))
    torch_keys = core_labels_to_torch(core_tensor.keys)
    return TorchTensorMap(torch_keys, torch_blocks)


def core_tensor_block_to_torch(core_block: TensorBlock):
    """Transforms a tensor block from metatensor-core to metatensor-torch
    :param core_block:
        tensor block from metatensor-core
    :returns torch_block:
        tensor block from metatensor-torch
    """
    block = TorchTensorBlock(
        values=torch.tensor(core_block.values),
        samples=core_labels_to_torch(core_block.samples),
        components=[
            core_labels_to_torch(component) for component in core_block.components
        ],
        properties=core_labels_to_torch(core_block.properties),
    )
    for parameter, gradient in core_block.gradients():
        block.add_gradient(
            parameter=parameter,
            gradient=TorchTensorBlock(
                values=torch.tensor(gradient.values),
                samples=core_labels_to_torch(gradient.samples),
                components=[
                    core_labels_to_torch(component) for component in gradient.components
                ],
                properties=core_labels_to_torch(gradient.properties),
            ),
        )
    return block


def core_labels_to_torch(core_labels: Labels):
    """Transforms labels from metatensor-core to metatensor-torch
    :param core_block:
        tensor block from metatensor-core
    :returns torch_block:
        labels from metatensor-torch
    """
    return TorchLabels(core_labels.names, torch.tensor(core_labels.values))


class SubsetOfRegressors:
    def __init__(
        self,
        kernel_kwargs: Optional[dict] = None,
    ):
        if kernel_kwargs is None:
            kernel_kwargs = {}

        # Set the kernel
        self._kernel: Union[AggregateKernel, None] = None
        self._kernel = AggregatePolynomial(**kernel_kwargs)

        self._kernel_kwargs = kernel_kwargs
        self._X_pseudo = None
        self._weights = None

    def fit(
        self,
        X: TensorMap,
        X_pseudo: TensorMap,
        y: TensorMap,
        alpha: float = 1.0,
        alpha_forces: Optional[float] = None,
    ):
        r"""
        :param X:
            features
            if kernel type "precomputed" is used, the kernel k_nm is assumed
        :param X_pseudo:
            pseudo points
            if kernel type "precomputed" is used, the kernel k_mm is assumed
        :param y:
            targets
        :param alpha:
            regularization for the energies, it must be a float
        :param alpha_forces:
            regularization for the forces, it must be a float. If None is set
            equal to alpha

        Derivation
        ----------

        We take equation the mean expression

        .. math::

            \sigma^{-2} K_{tm}\Sigma K_{MN}y

        we put in the $\Sigma$

        .. math::

            \sigma^{-2} K_{tm}(\sigma^{-2}K_{mn}K_{mn}+K_{mm})^{-1} K_{mn}y

        We can move around the $\sigma's$

        .. math::

             K_{tm}((K_{mn}\sigma^{-1})(\sigma^{-1}K_{mn)}+K_{mm})^{-1}
                            (K_{mn}\sigma^{-1})(y\sigma^{-1})

        you can see the building blocks in the code are $K_{mn}\sigma^{-1}$ and
        $y\sigma^{-1}$
        """
        if isinstance(alpha, float):
            alpha_energy = alpha
        else:
            raise ValueError("alpha must either be a float")

        if alpha_forces is None:
            alpha_forces = alpha_energy
        else:
            if not isinstance(alpha_forces, float):
                raise ValueError("alpha must either be a float")

        X = X.to(arrays="numpy")
        X_pseudo = X_pseudo.to(arrays="numpy")
        y = y.to(arrays="numpy")

        if self._kernel is None:
            # _set_kernel only returns None if kernel type is precomputed
            k_nm = X
            k_mm = X_pseudo
        else:
            k_mm = self._kernel(X_pseudo, X_pseudo, are_pseudo_points=(True, True))
            k_nm = self._kernel(X, X_pseudo, are_pseudo_points=(False, True))

        # solve
        # TODO: allow for different regularizer for energies and forces
        weight_blocks = []
        for key, y_block in y.items():
            k_nm_block = k_nm.block(key)
            k_mm_block = k_mm.block(key)
            X_block = X.block(key)
            structures = metatensor.operations._dispatch.unique(
                k_nm_block.samples["system"]
            )
            n_atoms_per_structure = []
            for structure in structures:
                n_atoms = np.sum(X_block.samples["system"] == structure)
                n_atoms_per_structure.append(float(n_atoms))

            n_atoms_per_structure = np.array(n_atoms_per_structure)
            normalization = metatensor.operations._dispatch.sqrt(n_atoms_per_structure)

            if not (np.allclose(alpha_energy, 0.0)):
                normalization /= alpha_energy
            normalization = normalization[:, None]

            k_nm_reg = k_nm_block.values * normalization
            y_reg = (y_block.values) * normalization
            if len(k_nm_block.gradients_list()) > 0:
                grad_shape = k_nm_block.gradient("positions").values.shape
                k_nm_reg = np.vstack(
                    [
                        k_nm_reg,
                        k_nm_block.gradient("positions").values.reshape(
                            grad_shape[0] * grad_shape[1],
                            grad_shape[2],
                        )
                        / alpha_forces,
                    ]
                )
                grad_shape = y_block.gradient("positions").values.shape
                y_reg = np.vstack(
                    [
                        y_reg,
                        y_block.gradient("positions").values.reshape(
                            grad_shape[0] * grad_shape[1],
                            grad_shape[2],
                        )
                        / alpha_forces,
                    ]
                )
            self._solver = _SorKernelSolver(k_mm_block.values)

            self._solver.fit(k_nm_reg, y_reg)

            weight_block = TensorBlock(
                values=self._solver.weights.T,
                samples=y_block.properties,
                components=k_nm_block.components,
                properties=k_nm_block.properties,
            )
            weight_blocks.append(weight_block)

        self._weights = TensorMap(y.keys, weight_blocks)

        self._X_pseudo = X_pseudo.copy()

    def predict(self, T: TensorMap) -> TensorMap:
        """
        :param T:
            features
            if kernel type "precomputed" is used, the kernel k_tm is assumed
        """
        if self._weights is None:
            raise ValueError(
                "The weights are not defined. Are you sure you"
                " have run the `fit` method?"
            )
        if self._kernel_type == "precomputed":
            k_tm = T
        else:
            k_tm = self._kernel(T, self._X_pseudo, are_pseudo_points=(False, True))
        return metatensor.dot(k_tm, self._weights)

    def export_torch_script_model(self):
        return TorchSubsetofRegressors(
            core_tensor_map_to_torch(self._weights),
            core_tensor_map_to_torch(self._X_pseudo),
            self._kernel_kwargs,
        )


class TorchSubsetofRegressors(torch.nn.Module):
    def __init__(
        self,
        weights: TorchTensorMap,
        X_pseudo: TorchTensorMap,
        kernel_kwargs: Optional[dict] = None,
    ):
        super().__init__()
        self._weights = weights
        self._X_pseudo = X_pseudo
        if kernel_kwargs is None:
            kernel_kwargs = {}

        # Set the kernel
        self._kernel = TorchAggregatePolynomial(**kernel_kwargs)

    def forward(self, T: TorchTensorMap) -> TorchTensorMap:
        """
        :param T:
            features
            if kernel type "precomputed" is used, the kernel k_tm is assumed
        """
        # move weights and X_pseudo to the same device as T
        self._weights = self._weights.to(T.device)
        self._X_pseudo = self._X_pseudo.to(T.device)

        k_tm = self._kernel(T, self._X_pseudo, are_pseudo_points=(False, True))
        return metatensor.torch.dot(k_tm, self._weights)



import torch


def cutoff_func(grid: torch.Tensor, r_cut: float, delta: float):
    mask_bigger = grid >= r_cut
    mask_smaller = grid <= r_cut - delta
    grid = (grid - r_cut + delta) / delta
    f = 0.5 + 0.5 * torch.cos(torch.pi * grid)

    f[mask_bigger] = 0.0
    f[mask_smaller] = 1.0
    return f


class DummyModule(torch.nn.Module):
    """Dummy torch module to make torchscript happy.
    This model should never be run"""

    def __init__(self):
        super(DummyModule, self).__init__()

    def forward(self, x) -> torch.Tensor:
        raise RuntimeError("This model should never be run")



import ase
from metatensor.torch.atomistic import System


def system_to_ase(system: System) -> ase.Atoms:
    """Converts a ``metatensor.torch.atomistic.System`` to an ``ase.Atoms`` object.
    This will discard any neighbor lists attached to the ``System``.

    :param system: The system to convert.

    :return: The system as an ``ase.Atoms`` object.
    """

    # Convert the system to an ASE atoms object
    positions = system.positions.detach().cpu().numpy()
    numbers = system.types.detach().cpu().numpy()
    cell = system.cell.detach().cpu().numpy()
    pbc = list(cell.any(axis=1))
    atoms = ase.Atoms(
        numbers=numbers,
        positions=positions,
        cell=cell,
        pbc=pbc,
    )

    return atoms



from typing import Dict, Optional, Tuple, Union

import torch
from metatensor.torch import TensorMap
from omegaconf import DictConfig
from torch.nn.modules.loss import _Loss

from metatrain.utils.external_naming import to_internal_name


class TensorMapLoss:
    """A loss function that operates on two ``metatensor.torch.TensorMap``.

    The loss is computed as the sum of the loss on the block values and
    the loss on the gradients, with weights specified at initialization.

    At the moment, this loss function assumes that all the gradients
    declared at initialization are present in both TensorMaps.

    :param reduction: The reduction to apply to the loss.
        See :py:class:`torch.nn.MSELoss`.
    :param weight: The weight to apply to the loss on the block values.
    :param gradient_weights: The weights to apply to the loss on the gradients.
    :param sliding_factor: The factor to apply to the exponential moving average
        of the "sliding" weights. These are weights that act on different components of
        the loss (for example, energies and forces), based on their individual recent
        history. If ``None``, no sliding weights are used in the computation of the
        loss.
    :param type: The type of loss to use. This can be either "mse" or "mae".
        A Huber loss can also be requested as a dictionary with the key "huber" and
        the value must be a dictionary with the key "deltas" and the value
        must be a dictionary with the keys "values" and the gradient keys.
        The values of the dictionary must be the deltas to use for the
        Huber loss.

    :returns: The loss as a zero-dimensional :py:class:`torch.Tensor`
        (with one entry).
    """

    def __init__(
        self,
        reduction: str = "mean",
        weight: float = 1.0,
        gradient_weights: Optional[Dict[str, float]] = None,
        sliding_factor: Optional[float] = None,
        type: Union[str, dict] = "mse",
    ):
        if gradient_weights is None:
            gradient_weights = {}

        losses = {}
        if type == "mse":
            losses["values"] = torch.nn.MSELoss(reduction=reduction)
            for key in gradient_weights.keys():
                losses[key] = torch.nn.MSELoss(reduction=reduction)
        elif type == "mae":
            losses["values"] = torch.nn.L1Loss(reduction=reduction)
            for key in gradient_weights.keys():
                losses[key] = torch.nn.L1Loss(reduction=reduction)
        elif isinstance(type, dict) and "huber" in type:
            # Huber loss
            deltas = type["huber"]["deltas"]
            losses["values"] = torch.nn.HuberLoss(
                reduction=reduction, delta=deltas["values"]
            )
            for key in gradient_weights.keys():
                losses[key] = torch.nn.HuberLoss(reduction=reduction, delta=deltas[key])
        else:
            raise ValueError(f"Unknown loss type: {type}")

        self.losses = losses
        self.weight = weight
        self.gradient_weights = gradient_weights
        self.sliding_factor = sliding_factor
        self.sliding_weights: Optional[Dict[str, TensorMap]] = None

    def __call__(
        self,
        predictions_tensor_map: TensorMap,
        targets_tensor_map: TensorMap,
    ) -> Tuple[torch.Tensor, Dict[str, Tuple[float, int]]]:
        # Check that the two have the same metadata, except for the samples,
        # which can be different due to batching, but must have the same size:
        if predictions_tensor_map.keys != targets_tensor_map.keys:
            raise ValueError(
                "TensorMapSlidingLoss requires the two TensorMaps to have the "
                "same keys."
            )
        for block_1, block_2 in zip(
            predictions_tensor_map.blocks(), targets_tensor_map.blocks()
        ):
            if block_1.properties != block_2.properties:
                raise ValueError(
                    "TensorMapSlidingLoss requires the two TensorMaps to have the same "
                    "properties."
                )
            if block_1.components != block_2.components:
                raise ValueError(
                    "TensorMapSlidingLoss requires the two TensorMaps to have the same "
                    "components."
                )
            if len(block_1.samples) != len(block_2.samples):
                raise ValueError(
                    "TensorMapSlidingLoss requires the two TensorMaps "
                    "to have the same number of samples."
                )
            for gradient_name in self.gradient_weights.keys():
                if len(block_1.gradient(gradient_name).samples) != len(
                    block_2.gradient(gradient_name).samples
                ):
                    raise ValueError(
                        "TensorMapSlidingLoss requires the two TensorMaps "
                        "to have the same number of gradient samples."
                    )
                if (
                    block_1.gradient(gradient_name).properties
                    != block_2.gradient(gradient_name).properties
                ):
                    raise ValueError(
                        "TensorMapSlidingLoss requires the two TensorMaps "
                        "to have the same gradient properties."
                    )
                if (
                    block_1.gradient(gradient_name).components
                    != block_2.gradient(gradient_name).components
                ):
                    raise ValueError(
                        "TensorMapSlidingLoss requires the two TensorMaps "
                        "to have the same gradient components."
                    )

        # First time the function is called: compute the sliding weights only
        # from the targets (if they are enabled)
        if self.sliding_factor is not None and self.sliding_weights is None:
            self.sliding_weights = get_sliding_weights(
                self.losses,
                self.sliding_factor,
                targets_tensor_map,
            )

        # Compute the loss:
        loss = torch.zeros(
            (),
            dtype=predictions_tensor_map.block(0).values.dtype,
            device=predictions_tensor_map.block(0).values.device,
        )
        for key in targets_tensor_map.keys:
            block_1 = predictions_tensor_map.block(key)
            block_2 = targets_tensor_map.block(key)
            values_1 = block_1.values
            values_2 = block_2.values
            # sliding weights: default to 1.0 if not used/provided for this target
            sliding_weight = (
                1.0
                if self.sliding_weights is None
                else self.sliding_weights.get("values", 1.0)
            )
            loss += (
                self.weight * self.losses["values"](values_1, values_2) / sliding_weight
            )
            for gradient_name, gradient_weight in self.gradient_weights.items():
                values_1 = block_1.gradient(gradient_name).values
                values_2 = block_2.gradient(gradient_name).values
                # sliding weights: default to 1.0 if not used/provided for this target
                sliding_weigths_value = (
                    1.0
                    if self.sliding_weights is None
                    else self.sliding_weights.get(gradient_name, 1.0)
                )
                loss += (
                    gradient_weight
                    * self.losses[gradient_name](values_1, values_2)
                    / sliding_weigths_value
                )
        if self.sliding_factor is not None:
            self.sliding_weights = get_sliding_weights(
                self.losses,
                self.sliding_factor,
                targets_tensor_map,
                predictions_tensor_map,
                self.sliding_weights,
            )
        return loss


class TensorMapDictLoss:
    """A loss function that operates on two ``Dict[str, metatensor.torch.TensorMap]``.

    At initialization, the user specifies a list of keys to use for the loss,
    along with a weight for each key.

    The loss is then computed as a weighted sum. Any keys that are not present
    in the dictionaries are ignored.

    :param weights: A dictionary mapping keys to weights. This might contain
        gradient keys, in the form ``<output_name>_<gradient_name>_gradients``.
    :param sliding_factor: The factor to apply to the exponential moving average
        of the "sliding" weights. These are weights that act on different components of
        the loss (for example, energies and forces), based on their individual recent
        history. If ``None``, no sliding weights are used in the computation of the
        loss.
    :param reduction: The reduction to apply to the loss.
        See :py:class:`torch.nn.MSELoss`.

    :returns: The loss as a zero-dimensional :py:class:`torch.Tensor`
        (with one entry).
    """

    def __init__(
        self,
        weights: Dict[str, float],
        sliding_factor: Optional[float] = None,
        reduction: str = "mean",
        type: Union[str, dict] = "mse",
    ):
        outputs = [key for key in weights.keys() if "gradients" not in key]
        self.losses = {}
        for output in outputs:
            value_weight = weights[output]
            gradient_weights = {}
            for key, weight in weights.items():
                if key.startswith(output) and key.endswith("_gradients"):
                    gradient_name = key.replace(f"{output}_", "").replace(
                        "_gradients", ""
                    )
                    gradient_weights[gradient_name] = weight
            type_output = _process_type(type, output)
            if output == "energy" and sliding_factor is not None:
                self.losses[output] = TensorMapLoss(
                    reduction=reduction,
                    weight=value_weight,
                    gradient_weights=gradient_weights,
                    sliding_factor=sliding_factor,
                    type=type_output,
                )
            else:
                self.losses[output] = TensorMapLoss(
                    reduction=reduction,
                    weight=value_weight,
                    gradient_weights=gradient_weights,
                    type=type_output,
                )

    def __call__(
        self,
        tensor_map_dict_1: Dict[str, TensorMap],
        tensor_map_dict_2: Dict[str, TensorMap],
    ) -> torch.Tensor:
        # Assert that the two have the keys:
        assert set(tensor_map_dict_1.keys()) == set(tensor_map_dict_2.keys())

        # Initialize the loss:
        first_values = next(iter(tensor_map_dict_1.values())).block(0).values
        loss = torch.zeros((), dtype=first_values.dtype, device=first_values.device)

        # Compute the loss:
        for target in tensor_map_dict_1.keys():
            target_loss = self.losses[target](
                tensor_map_dict_1[target], tensor_map_dict_2[target]
            )
            loss += target_loss

        return loss


def get_sliding_weights(
    losses: Dict[str, _Loss],
    sliding_factor: float,
    targets: TensorMap,
    predictions: Optional[TensorMap] = None,
    previous_sliding_weights: Optional[Dict[str, float]] = None,
) -> Dict[str, float]:
    """
    Compute the sliding weights for the loss function.

    The sliding weights are computed as the absolute difference between the
    predictions and the targets.

    :param predictions: The predictions.
    :param targets: The targets.

    :return: The sliding weights.
    """
    sliding_weights = {}
    if predictions is None:
        for block in targets.blocks():
            values = block.values
            sliding_weights["values"] = (
                losses["values"](values, values.mean() * torch.ones_like(values)) + 1e-6
            )
            for gradient_name, gradient_block in block.gradients():
                values = gradient_block.values
                sliding_weights[gradient_name] = losses[gradient_name](
                    values, torch.zeros_like(values)
                )
    elif predictions is not None:
        if previous_sliding_weights is None:
            raise RuntimeError(
                "previous_sliding_weights must be provided if predictions is not None"
            )
        else:
            for predictions_block, target_block in zip(
                predictions.blocks(), targets.blocks()
            ):
                target_values = target_block.values
                predictions_values = predictions_block.values
                sliding_weights["values"] = (
                    sliding_factor * previous_sliding_weights["values"]
                    + (1 - sliding_factor)
                    * losses["values"](predictions_values, target_values).detach()
                )
                for gradient_name, gradient_block in target_block.gradients():
                    target_values = gradient_block.values
                    predictions_values = predictions_block.gradient(
                        gradient_name
                    ).values
                    sliding_weights[gradient_name] = (
                        sliding_factor * previous_sliding_weights[gradient_name]
                        + (1 - sliding_factor)
                        * losses[gradient_name](
                            predictions_values, target_values
                        ).detach()
                    )
    return sliding_weights


def _process_type(type: Union[str, DictConfig], output: str) -> Union[str, dict]:
    if not isinstance(type, str):
        assert "huber" in type
        # we process the Huber loss delta dict to make it similar to the
        # `weights` dict
        type_output = {"huber": {"deltas": {}}}  # type: ignore
        for key, delta in type["huber"]["deltas"].items():
            key_internal = to_internal_name(key)
            if key_internal == output:
                type_output["huber"]["deltas"]["values"] = delta
            elif key_internal.startswith(output) and key_internal.endswith(
                "_gradients"
            ):
                gradient_name = key_internal.replace(f"{output}_", "").replace(
                    "_gradients", ""
                )
                type_output["huber"]["deltas"][gradient_name] = delta
            else:
                pass
    else:
        type_output = type  # type: ignore
    return type_output



from typing import Any, Dict


def update_hypers(
    hypers: Dict[str, Any], model_hypers: Dict[str, Any], do_forces: bool = True
):
    """
    Updates the hypers dictionary with the model hypers, the
    MLIP_SETTINGS and UTILITY_FLAGS keys of the PET model.
    """

    # set model hypers
    hypers = hypers.copy()
    hypers["ARCHITECTURAL_HYPERS"] = model_hypers
    hypers["ARCHITECTURAL_HYPERS"]["DTYPE"] = "float32"
    hypers["ARCHITECTURAL_HYPERS"]["D_OUTPUT"] = 1
    hypers["ARCHITECTURAL_HYPERS"]["TARGET_TYPE"] = "structural"
    hypers["ARCHITECTURAL_HYPERS"]["TARGET_AGGREGATION"] = "sum"

    # set MLIP_SETTINGS
    hypers["MLIP_SETTINGS"] = {
        "ENERGY_KEY": "energy",
        "FORCES_KEY": "forces",
        "USE_ENERGIES": True,
        "USE_FORCES": do_forces,
    }

    # set PET utility flags
    hypers["UTILITY_FLAGS"] = {
        "CALCULATION_TYPE": None,
    }
    return hypers



from typing import List

import torch
from metatensor.torch import Labels, TensorBlock, TensorMap


@torch.jit.script
def sum_over_atoms(tensor_map: TensorMap):  # pragma: no cover
    """
    A faster version of ``metatensor.torch.sum_over_samples``, specialized for
    summing over atoms in graph-like TensorMaps.

    :param tensor_map: The TensorMap to sum over.
    :return: A new TensorMap with the same keys, but with the samples summed
        over the atoms.
    """
    new_blocks: List[TensorBlock] = []
    for block in tensor_map.blocks():
        n_systems = int(block.samples.column("system").max() + 1)
        new_tensor = torch.zeros(
            [n_systems] + block.values.shape[1:],
            device=tensor_map.device,
            dtype=tensor_map.dtype,
        )
        new_tensor.index_add_(0, block.samples.column("system"), block.values)
        new_block = TensorBlock(
            values=new_tensor,
            samples=Labels(
                names=["system"],
                values=torch.arange(
                    n_systems, device=tensor_map.device, dtype=torch.int
                ).reshape(-1, 1),
            ),
            components=block.components,
            properties=block.properties,
        )
        new_blocks.append(new_block)
    return TensorMap(
        keys=tensor_map.keys,
        blocks=new_blocks,
    )



import copy
import random
import re

import numpy as np
import pytest
import torch
from omegaconf import OmegaConf

from metatrain.gap import GAP, Trainer
from metatrain.utils.data import Dataset, DatasetInfo, read_systems, read_targets
from metatrain.utils.data.target_info import (
    get_energy_target_info,
    get_generic_target_info,
)

from . import DATASET_ETHANOL_PATH, DEFAULT_HYPERS


torch.set_default_dtype(torch.float64)  # GAP only supports float64


# reproducibility
random.seed(0)
np.random.seed(0)
torch.manual_seed(0)


def test_ethanol_regression_train_and_invariance():
    """test the error if the number of sparse point
    is bigger than the number of environments
    """

    systems = read_systems(DATASET_ETHANOL_PATH)

    conf = {
        "energy": {
            "quantity": "energy",
            "read_from": DATASET_ETHANOL_PATH,
            "reader": "ase",
            "key": "energy",
            "unit": "kcal/mol",
            "type": "scalar",
            "per_atom": False,
            "num_subtargets": 1,
            "forces": {
                "read_from": DATASET_ETHANOL_PATH,
                "reader": "ase",
                "key": "forces",
            },
            "stress": False,
            "virial": False,
        }
    }

    targets, _ = read_targets(OmegaConf.create(conf))
    dataset = Dataset.from_dict(
        {"system": systems[:2], "energy": targets["energy"][:2]}
    )

    hypers = copy.deepcopy(DEFAULT_HYPERS)
    hypers["model"]["krr"]["num_sparse_points"] = 30

    target_info_dict = {
        "energy": get_energy_target_info(
            {"unit": "kcal/mol"}, add_position_gradients=True
        )
    }

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[1, 6, 7, 8], targets=target_info_dict
    )

    gap = GAP(hypers["model"], dataset_info)
    trainer = Trainer(hypers["training"])
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Number of sparse points (30) "
            "should be smaller than the number of environments (18)"
        ),
    ):
        trainer.train(
            model=gap,
            dtype=torch.float64,
            devices=[torch.device("cpu")],
            train_datasets=[dataset],
            val_datasets=[dataset],
            checkpoint_dir=".",
        )


@pytest.mark.parametrize("per_atom", [True, False])
def test_vector_output(per_atom):
    """Tests that the model can predict a (spherical) vector output."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "forces": get_generic_target_info(
                {
                    "quantity": "forces",
                    "unit": "",
                    "type": {
                        "spherical": {"irreps": [{"o3_lambda": 1, "o3_sigma": 1}]}
                    },
                    "num_subtargets": 100,
                    "per_atom": per_atom,
                }
            )
        },
    )

    with pytest.raises(ValueError, match="GAP only supports total-energy-like outputs"):
        GAP(DEFAULT_HYPERS["model"], dataset_info)



import copy
import random

import metatensor.torch
import numpy as np
import torch
from omegaconf import OmegaConf

from metatrain.gap import GAP, Trainer
from metatrain.utils.data import Dataset, DatasetInfo
from metatrain.utils.data.readers import read_systems, read_targets
from metatrain.utils.data.readers.ase import read
from metatrain.utils.data.target_info import get_energy_target_info

from . import DATASET_ETHANOL_PATH, DATASET_PATH, DEFAULT_HYPERS


torch.set_default_dtype(torch.float64)  # GAP only supports float64


# reproducibility
random.seed(0)
np.random.seed(0)
torch.manual_seed(0)


def test_regression_init():
    """Perform a regression test on the model at initialization"""
    targets = {}
    targets["mtt::U0"] = get_energy_target_info({"unit": "eV"})

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[1, 6, 7, 8], targets=targets
    )
    GAP(DEFAULT_HYPERS["model"], dataset_info)


def test_regression_train_and_invariance():
    """Perform a regression test on the model when trained for 2 epoch on a small
    dataset.  We perform also the invariance test here because one needs a trained model
    for this.
    """

    systems = read_systems(DATASET_PATH)

    conf = {
        "mtt::U0": {
            "quantity": "energy",
            "read_from": DATASET_PATH,
            "reader": "ase",
            "key": "U0",
            "unit": "kcal/mol",
            "type": "scalar",
            "per_atom": False,
            "num_subtargets": 1,
            "forces": False,
            "stress": False,
            "virial": False,
        }
    }
    targets, _ = read_targets(OmegaConf.create(conf))
    dataset = Dataset.from_dict({"system": systems, "mtt::U0": targets["mtt::U0"]})

    target_info_dict = {}
    target_info_dict["mtt::U0"] = get_energy_target_info({"unit": "eV"})

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[1, 6, 7, 8], targets=target_info_dict
    )

    gap = GAP(DEFAULT_HYPERS["model"], dataset_info)
    trainer = Trainer(DEFAULT_HYPERS["training"])
    trainer.train(
        model=gap,
        dtype=torch.float64,
        devices=[torch.device("cpu")],
        train_datasets=[dataset],
        val_datasets=[dataset],
        checkpoint_dir=".",
    )
    gap.eval()

    # Predict on the first five systems
    output = gap(systems[:5], {"mtt::U0": gap.outputs["mtt::U0"]})

    expected_output = torch.tensor(
        [[-40.5891], [-56.7122], [-76.4146], [-77.3364], [-93.4905]]
    )

    assert torch.allclose(output["mtt::U0"].block().values, expected_output, rtol=0.3)

    # Tests that the model is rotationally invariant
    system = read(DATASET_PATH)
    system.numbers = np.ones(len(system.numbers))

    original_system = copy.deepcopy(system)
    system.rotate(48, "y")

    original_output = gap(
        [metatensor.torch.atomistic.systems_to_torch(original_system)],
        {"mtt::U0": gap.outputs["mtt::U0"]},
    )
    rotated_output = gap(
        [metatensor.torch.atomistic.systems_to_torch(system)],
        {"mtt::U0": gap.outputs["mtt::U0"]},
    )

    assert torch.allclose(
        original_output["mtt::U0"].block().values,
        rotated_output["mtt::U0"].block().values,
    )


def test_ethanol_regression_train_and_invariance():
    """Perform a regression test on the model when trained for 2 epoch on a small
    dataset.  We perform also the invariance test here because one needs a trained model
    for this.
    """

    systems = read_systems(DATASET_ETHANOL_PATH)

    conf = {
        "energy": {
            "quantity": "energy",
            "read_from": DATASET_ETHANOL_PATH,
            "reader": "ase",
            "key": "energy",
            "type": "scalar",
            "per_atom": False,
            "num_subtargets": 1,
            "forces": {
                "read_from": DATASET_ETHANOL_PATH,
                "reader": "ase",
                "key": "forces",
            },
            "unit": "kcal/mol",
            "stress": False,
            "virial": False,
        }
    }

    targets, _ = read_targets(OmegaConf.create(conf))
    dataset = Dataset.from_dict({"system": systems, "energy": targets["energy"]})

    hypers = copy.deepcopy(DEFAULT_HYPERS)
    hypers["model"]["krr"]["num_sparse_points"] = 900

    target_info_dict = {
        "energy": get_energy_target_info({"unit": "eV"}, add_position_gradients=True)
    }

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[1, 6, 7, 8], targets=target_info_dict
    )

    gap = GAP(hypers["model"], dataset_info)
    trainer = Trainer(hypers["training"])
    trainer.train(
        model=gap,
        dtype=torch.float64,
        devices=[torch.device("cpu")],
        train_datasets=[dataset],
        val_datasets=[dataset],
        checkpoint_dir=".",
    )
    gap.eval()

    # Predict on the first five systems
    output = gap(systems[:5], {"energy": gap.outputs["energy"]})
    data = read(DATASET_ETHANOL_PATH, ":5", format="extxyz")

    expected_output = torch.tensor([[i.info["energy"]] for i in data])
    assert torch.allclose(output["energy"].block().values, expected_output, rtol=0.1)

    # TODO: check accuracy of training forces
    # expected_forces = torch.vstack([torch.Tensor(i.arrays["forces"]) for i in data])

    # Tests that the model is rotationally invariant
    system = read(DATASET_ETHANOL_PATH)

    original_system = copy.deepcopy(system)
    system.rotate(48, "y")

    original_output = gap(
        [metatensor.torch.atomistic.systems_to_torch(original_system)],
        {"energy": gap.outputs["energy"]},
    )
    rotated_output = gap(
        [metatensor.torch.atomistic.systems_to_torch(system)],
        {"energy": gap.outputs["energy"]},
    )

    assert torch.allclose(
        original_output["energy"].block().values,
        rotated_output["energy"].block().values,
    )



import copy

import torch
from metatensor.torch.atomistic import System

from metatrain.experimental.nanopet import NanoPET
from metatrain.utils.data import DatasetInfo
from metatrain.utils.data.target_info import get_energy_target_info
from metatrain.utils.neighbor_lists import get_system_with_neighbor_lists

from . import MODEL_HYPERS


def test_torchscript():
    """Tests that the model can be jitted."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "energy": get_energy_target_info({"quantity": "energy", "unit": "eV"})
        },
    )
    model = NanoPET(MODEL_HYPERS, dataset_info)

    system = System(
        types=torch.tensor([6, 1, 8, 7]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]]
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    system = get_system_with_neighbor_lists(system, model.requested_neighbor_lists())

    model = torch.jit.script(model)
    model(
        [system],
        {"energy": model.outputs["energy"]},
    )


def test_torchscript_save_load(tmpdir):
    """Tests that the model can be jitted and saved."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "energy": get_energy_target_info({"quantity": "energy", "unit": "eV"})
        },
    )
    model = NanoPET(MODEL_HYPERS, dataset_info)

    with tmpdir.as_cwd():
        torch.jit.save(torch.jit.script(model), "model.pt")
        torch.jit.load("model.pt")


def test_torchscript_integers():
    """Tests that the model can be jitted when some float
    parameters are instead supplied as integers."""

    new_hypers = copy.deepcopy(MODEL_HYPERS)
    new_hypers["cutoff"] = 5
    new_hypers["cutoff_width"] = 1

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "energy": get_energy_target_info({"quantity": "energy", "unit": "eV"})
        },
    )
    model = NanoPET(new_hypers, dataset_info)

    system = System(
        types=torch.tensor([6, 1, 8, 7]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]]
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    system = get_system_with_neighbor_lists(system, model.requested_neighbor_lists())

    model = torch.jit.script(model)
    model(
        [system],
        {"energy": model.outputs["energy"]},
    )



import difflib
import importlib
import json
import logging
from importlib.util import find_spec
from pathlib import Path
from typing import Dict, List, Union

from omegaconf import OmegaConf

from .. import PACKAGE_ROOT
from .jsonschema import validate


def check_architecture_name(name: str) -> None:
    """Check if the requested architecture is available.

    If the architecture is not found an :func:`ValueError` is raised. If an architecture
    with the same name as an experimental or deprecated architecture exist, this
    architecture is suggested. If no architecture exist the closest architecture is
    given to help debugging typos.

    :param name: name of the architecture
    :raises ValueError: if the architecture is not found
    """
    try:
        if find_spec(f"metatrain.{name}") is not None:
            return
        elif find_spec(f"metatrain.experimental.{name}") is not None:
            msg = (
                f"Architecture {name!r} is not a stable architecture. An "
                "experimental architecture with the same name was found. Set "
                f"`name: experimental.{name}` in your options file to use this "
                "experimental architecture."
            )
        elif find_spec(f"metatrain.deprecated.{name}") is not None:
            msg = (
                f"Architecture {name!r} is not a stable architecture. A "
                "deprecated architecture with the same name was found. Set "
                f"`name: deprecated.{name}` in your options file to use this "
                "deprecated architecture."
            )
    except ModuleNotFoundError:
        msg = f"Architecture {name!r} is not a valid architecture."

        closest_match = difflib.get_close_matches(
            word=name, possibilities=find_all_architectures()
        )
        if closest_match:
            msg += f" Do you mean '{closest_match[0]}'?"

    raise ValueError(msg)


def check_architecture_options(
    name: str,
    options: Dict,
) -> None:
    """Verifies that an options instance only contains valid keys

    If the architecture developer does not provide a validation scheme the ``options``
    will not checked.

    :param name: name of the architecture
    :param options: architecture options to check
    """
    schema_path = get_architecture_path(name) / "schema-hypers.json"
    if schema_path.exists():
        with open(schema_path, "r") as f:
            schema = json.load(f)

        validate(instance=options, schema=schema)
    else:
        logging.debug("No schema found for {name!r} architecture. Skipping validation.")


def get_architecture_name(path: Union[str, Path]) -> str:
    """Name of an architecture based on path to pointing inside an architecture.

    The function should be used to determine the ``ARCHITECTURE_NAME`` based on the name
    of the folder.

    :param absolute_architecture_path: absolute path of the architecture directory
    :returns: architecture name
    :raises ValueError: if ``absolute_architecture_path`` does not point to a valid
        architecture directory.

    .. seealso::
        :py:func:`get_architecture_path` to get the relative path within the metatrain
        project of an architecture name.
    """
    path = Path(path)

    if path.is_dir():
        directory = path
    elif path.is_file():
        directory = path.parent
    else:
        raise ValueError(f"`path` {str(path)!r} does not exist")

    architecture_path = directory.relative_to(PACKAGE_ROOT)
    name = str(architecture_path).replace("/", ".")

    try:
        check_architecture_name(name)
    except ValueError as err:
        raise ValueError(
            f"`path` {str(path)!r} does not point to a valid architecture folder"
        ) from err

    return name


def import_architecture(name: str):
    """Import an architecture.

    :param name: name of the architecture
    :raises ImportError: if the architecture dependencies are not met
    """
    check_architecture_name(name)
    try:
        return importlib.import_module(f"metatrain.{name}")
    except ImportError as err:
        # consistent name with pyproject.toml's `optional-dependencies` section
        name_for_deps = name
        if "experimental." in name or "deprecated." in name:
            name_for_deps = ".".join(name.split(".")[1:])

        name_for_deps = name_for_deps.replace("_", "-")

        raise ImportError(
            f"Trying to import '{name}' but architecture dependencies "
            f"seem not be installed. \n"
            f"Try to install them with `pip install metatrain[{name_for_deps}]`"
        ) from err


def get_architecture_path(name: str) -> Path:
    """Return the relative path to the architecture directory.

    Path based on the ``name`` within the metatrain project directory.

    :param name: name of the architecture
    :returns: path to the architecture directory

    .. seealso::
        :py:func:`get_architecture_name` to get the name based on an absolute path of an
        architecture.
    """
    check_architecture_name(name)
    return PACKAGE_ROOT / Path(name.replace(".", "/"))


def find_all_architectures() -> List[str]:
    """Find all currently available architectures.

    To find the architectures the function searches for the mandatory
    ``default-hypers.yaml`` file in each architecture directory.

    :returns: List of architectures names
    """
    options_files_path = PACKAGE_ROOT.rglob("default-hypers.yaml")

    architecture_names = []
    for option_file_path in options_files_path:
        architecture_names.append(get_architecture_name(option_file_path))

    return architecture_names


def get_default_hypers(name: str) -> Dict:
    """Dictionary of the default architecture hyperparameters.

    :param: name of the architecture
    :returns: default hyper parameters of the architectures
    """
    check_architecture_name(name)
    default_hypers = OmegaConf.load(get_architecture_path(name) / "default-hypers.yaml")
    # We present the `default-hypers.yaml` file inside the documentation. For a better
    # user experience we store these yaml files with an additional level of indentation
    # (`"architecture"`), which we have to remove here to get the raw default hypers.
    return OmegaConf.to_container(default_hypers)["architecture"]



import copy

import pytest
import torch
from metatensor.torch.atomistic import System

from metatrain.soap_bpnn import SoapBpnn
from metatrain.utils.data import DatasetInfo
from metatrain.utils.data.target_info import (
    get_energy_target_info,
    get_generic_target_info,
)

from . import MODEL_HYPERS


def test_torchscript():
    """Tests that the model can be jitted."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    model = SoapBpnn(MODEL_HYPERS, dataset_info)
    model = torch.jit.script(model)

    system = System(
        types=torch.tensor([6, 1, 8, 7]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]]
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    model(
        [system],
        {"energy": model.outputs["energy"]},
    )


def test_torchscript_with_identity():
    """Tests that the model can be jitted."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    hypers = copy.deepcopy(MODEL_HYPERS)
    hypers["bpnn"]["layernorm"] = False
    model = SoapBpnn(hypers, dataset_info)
    model = torch.jit.script(model)

    system = System(
        types=torch.tensor([6, 1, 8, 7]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]]
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    model(
        [system],
        {"energy": model.outputs["energy"]},
    )


@pytest.mark.parametrize("o3_lambda", [0, 1, 2, 3])
@pytest.mark.parametrize("o3_sigma", [-1, 1])
def test_torchscript_spherical(o3_lambda, o3_sigma):
    """Tests that the spherical modules can be jitted."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "spherical_target": get_generic_target_info(
                {
                    "quantity": "",
                    "unit": "",
                    "type": {
                        "spherical": {
                            "irreps": [{"o3_lambda": o3_lambda, "o3_sigma": o3_sigma}]
                        }
                    },
                    "num_subtargets": 100,
                    "per_atom": False,
                }
            )
        },
    )
    model = SoapBpnn(MODEL_HYPERS, dataset_info)
    model = torch.jit.script(model)

    system = System(
        types=torch.tensor([6, 1, 8, 7]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]]
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    model(
        [system],
        {"spherical_target": model.outputs["spherical_target"]},
    )


def test_torchscript_save_load(tmpdir):
    """Tests that the model can be jitted and saved."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    model = SoapBpnn(MODEL_HYPERS, dataset_info)

    with tmpdir.as_cwd():
        torch.jit.save(torch.jit.script(model), "model.pt")
        torch.jit.load("model.pt")


def test_torchscript_integers():
    """Tests that the model can be jitted when some float
    parameters are instead supplied as integers."""

    new_hypers = copy.deepcopy(MODEL_HYPERS)
    new_hypers["soap"]["cutoff"]["radius"] = 5
    new_hypers["soap"]["density"]["width"] = 1
    new_hypers["soap"]["density"]["center_atom_weight"] = 1
    new_hypers["soap"]["cutoff"]["smoothing"]["width"] = 1
    new_hypers["soap"]["density"]["scaling"]["rate"] = 1
    new_hypers["soap"]["density"]["scaling"]["scale"] = 2
    new_hypers["soap"]["density"]["scaling"]["exponent"] = 7

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    model = SoapBpnn(new_hypers, dataset_info)
    model = torch.jit.script(model)

    system = System(
        types=torch.tensor([6, 1, 8, 7]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]]
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    model(
        [system],
        {"energy": model.outputs["energy"]},
    )



import torch


class AttentionBlock(torch.nn.Module):
    """
    A single transformer attention block. We are not using the
    MultiHeadAttention module from torch.nn because we need to apply a
    radial mask to the attention weights.
    """

    def __init__(
        self,
        hidden_size: int,
        num_heads: int,
        dropout_rate: float,
        attention_dropout_rate: float,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.in_proj = torch.nn.Linear(hidden_size, 3 * hidden_size, bias=False)
        self.out_proj = torch.nn.Linear(hidden_size, hidden_size, bias=False)
        self.layernorm = torch.nn.LayerNorm(normalized_shape=hidden_size)
        self.attention_dropout_rate = attention_dropout_rate

    def forward(
        self,
        inputs: torch.Tensor,  # seq_len hidden_size
        radial_mask: torch.Tensor,  # seq_len
    ) -> torch.Tensor:  # seq_len hidden_size
        # Pre-layer normalization
        normed_inputs = self.layernorm(inputs)

        # Input projection
        qkv = self.in_proj(normed_inputs)
        q, k, v = torch.chunk(qkv, 3, dim=-1)
        # Split heads
        q = q.reshape(q.size(0), q.size(1), self.num_heads, q.size(2) // self.num_heads)
        k = k.reshape(k.size(0), k.size(1), self.num_heads, k.size(2) // self.num_heads)
        v = v.reshape(v.size(0), v.size(1), self.num_heads, v.size(2) // self.num_heads)
        q = q.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)
        # Attention
        attention_weights = torch.matmul(q, k.transpose(-2, -1)) / (k.size(-1) ** 0.5)
        attention_weights = attention_weights.softmax(dim=-1)
        attention_weights = torch.nn.functional.dropout(
            attention_weights, p=self.attention_dropout_rate, training=self.training
        )

        # Radial mask
        attention_weights = attention_weights * radial_mask[:, None, None, :]
        attention_weights = attention_weights / (
            attention_weights.sum(dim=-1, keepdim=True) + 1e-6
        )

        # Attention output
        attention_output = torch.matmul(attention_weights, v)
        attention_output = attention_output.transpose(1, 2)
        attention_output = attention_output.reshape(
            attention_output.size(0),
            attention_output.size(1),
            attention_output.size(2) * attention_output.size(3),
        )

        # Output projection
        outputs = self.out_proj(attention_output)

        # Residual connection
        outputs = (outputs + inputs) * 0.5**0.5

        return outputs



import logging
from typing import Dict, List, Optional, Union

import metatensor.torch
import torch
from metatensor.torch import Labels, LabelsEntry, TensorBlock, TensorMap
from metatensor.torch.atomistic import ModelOutput, System

from ..data import Dataset, DatasetInfo, TargetInfo, get_all_targets, get_atomic_types
from ..jsonschema import validate
from ..sum_over_atoms import sum_over_atoms
from ..transfer import systems_and_targets_to_device
from .remove import remove_additive


class CompositionModel(torch.nn.Module):
    """A simple model that calculates the contributions to scalar targets
    based on the stoichiometry in a system.

    :param model_hypers: A dictionary of model hyperparameters. The paramater is ignored
        and is only present to be consistent with the general model API.
    :param dataset_info: An object containing information about the dataset, including
        target quantities and atomic types.
    """

    weights: Dict[str, TensorMap]
    outputs: Dict[str, ModelOutput]

    def __init__(self, model_hypers: Dict, dataset_info: DatasetInfo):
        super().__init__()

        # `model_hypers` should be an empty dictionary
        validate(
            instance=model_hypers,
            schema={"type": "object", "additionalProperties": False},
        )

        self.dataset_info = dataset_info
        self.atomic_types = sorted(dataset_info.atomic_types)

        self.register_buffer(
            "type_to_index", torch.empty(max(self.atomic_types) + 1, dtype=torch.long)
        )
        for i, atomic_type in enumerate(self.atomic_types):
            self.type_to_index[atomic_type] = i

        for target_name, target_info in dataset_info.targets.items():
            if not self.is_valid_target(target_name, target_info):
                raise ValueError(
                    f"Composition model does not support target quantity "
                    f"{target_info.quantity}. This is an architecture bug. "
                    "Please report this issue and help us improve!"
                )

        self.new_targets = {
            target_name: target_info
            for target_name, target_info in dataset_info.targets.items()
        }

        self.weights = {}
        self.outputs: Dict[str, ModelOutput] = {}
        for target_name, target_info in self.dataset_info.targets.items():
            self._add_output(target_name, target_info)

        # keeps track of dtype and device of the composition model
        self.register_buffer("dummy_buffer", torch.randn(1))

    def train_model(
        self,
        datasets: List[Union[Dataset, torch.utils.data.Subset]],
        additive_models: List[torch.nn.Module],
        fixed_weights: Optional[Dict[str, Dict[int, str]]] = None,
    ) -> None:
        """Train/fit the composition weights for the datasets.

        :param datasets: Dataset(s) to calculate the composition weights for.
        :param fixed_weights: Optional fixed weights to use for the composition model,
            for one or more target quantities.
        :param additive_models: Additive models to be removed from the targets
            before calculating the statistics.

        :raises ValueError: If the provided datasets contain unknown targets.
        :raises ValueError: If the provided datasets contain unknown atomic types.
        :raises RuntimeError: If the linear system to calculate the composition weights
            cannot be solved.
        """
        if not isinstance(datasets, list):
            datasets = [datasets]

        if fixed_weights is None:
            fixed_weights = {}

        additional_types = sorted(
            set(get_atomic_types(datasets)) - set(self.atomic_types)
        )
        if additional_types:
            raise ValueError(
                "Provided `datasets` contains unknown "
                f"atomic types {additional_types}. "
                f"Known types from initialization are {self.atomic_types}."
            )

        missing_types = sorted(set(self.atomic_types) - set(get_atomic_types(datasets)))
        if missing_types:
            logging.warning(
                f"Provided `datasets` do not contain atomic types {missing_types}. "
                f"Known types from initialization are {self.atomic_types}."
            )

        device = self.dummy_buffer.device

        # Fill the weights for each "new" target (i.e. those that do not already
        # have composition weights from a previous training run)
        for target_key in self.new_targets:
            if target_key in fixed_weights:
                if not self.dataset_info.targets[target_key].is_scalar:
                    raise ValueError(
                        "Fixed weights can only be provided for scalar targets. "
                        f"Target {target_key} is not scalar."
                    )
                if (
                    len(self.dataset_info.targets[target_key].layout.block().properties)
                    != 1
                ):
                    raise ValueError(
                        "Fixed weights can only be provided for targets with one "
                        f"property. Target {target_key} has more than one property."
                    )
                # The fixed weights are provided for this target. Use them:
                if not sorted(fixed_weights[target_key].keys()) == self.atomic_types:
                    raise ValueError(
                        f"Fixed weights for target {target_key} must contain all "
                        f"atomic types {self.atomic_types}."
                    )
                weights_tensor = torch.tensor(
                    [fixed_weights[target_key][i] for i in self.atomic_types],
                    dtype=self.dummy_buffer.dtype,
                    device=device,
                ).reshape(-1, 1)
                self.weights[target_key] = TensorMap(
                    keys=Labels.single().to(device),
                    blocks=[
                        TensorBlock(
                            values=weights_tensor,
                            samples=Labels(
                                names=["center_type"],
                                values=torch.tensor(
                                    self.atomic_types, dtype=torch.int, device=device
                                ).reshape(-1, 1),
                            ),
                            components=[
                                c.to(device)
                                for c in self.dataset_info.targets[target_key]
                                .layout.block()
                                .components
                            ],
                            properties=self.dataset_info.targets[target_key]
                            .layout.block()
                            .properties.to(device),
                        )
                    ],
                )
            else:
                datasets_with_target = []
                for dataset in datasets:
                    if target_key in get_all_targets(dataset):
                        datasets_with_target.append(dataset)
                if len(datasets_with_target) == 0:
                    # this is a possibility when transfer learning
                    logging.warning(
                        f"Target {target_key} in the model's new capabilities is not "
                        "present in any of the training datasets."
                    )
                    continue

                total_num_structures = sum(
                    [len(dataset) for dataset in datasets_with_target]
                )
                dtype = datasets[0][0]["system"].positions.dtype
                if dtype != torch.float64:
                    raise ValueError(
                        "The composition model only supports float64 during training. "
                        f"Got dtype: {dtype}."
                    )

                composition_features = torch.zeros(
                    (total_num_structures, len(self.atomic_types)),
                    dtype=dtype,
                    device=device,
                )
                system_index = 0
                per_block_targets_list: Dict[LabelsEntry, List[TensorBlock]] = {}
                for dataset in datasets_with_target:
                    for sample in dataset:
                        systems = [sample["system"]]
                        targets = {target_key: sample[target_key]}
                        systems, targets = systems_and_targets_to_device(
                            systems, targets, device
                        )
                        for additive_model in additive_models:
                            target_info_dict = {
                                target_key: self.new_targets[target_key]
                            }
                            targets = remove_additive(  # remove other additive models
                                systems,
                                targets,
                                additive_model,
                                target_info_dict,
                            )
                        for j, t in enumerate(self.atomic_types):
                            composition_features[system_index, j] = torch.sum(
                                systems[0].types == t
                            )
                        system_index += 1
                        if self.dataset_info.targets[target_key].per_atom:
                            # we need the center type in the samples to do
                            # mean_over_samples
                            if "center_type" in targets[target_key].keys.names:
                                # it's in the keys: move it to the samples
                                targets[target_key] = targets[
                                    target_key
                                ].keys_to_samples("center_type")
                            if targets[target_key].block(0).samples.names == [
                                "system",
                                "atom",
                            ]:
                                # there is no center type, we need to add it
                                # and we will rely on the fact that per-atom targets
                                # should be in the same order as the atoms in the system
                                targets[target_key] = metatensor.torch.append_dimension(
                                    targets[target_key],
                                    "samples",
                                    "center_type",
                                    systems[0].types,
                                )
                        # TODO: abstract even more for more complex targets?
                        for key, block in targets[target_key].items():
                            # `if key not in per_block_targets_list` doesn't work, so:
                            matching_keys = [
                                k for k in per_block_targets_list if k == key
                            ]
                            assert len(matching_keys) <= 1
                            if len(matching_keys) == 0:
                                per_block_targets_list[key] = [block]
                            else:
                                per_block_targets_list[matching_keys[0]].append(block)

                weight_blocks = []
                for key, block_list in per_block_targets_list.items():
                    # distinguish between spherical and scalar targets
                    needs_unsqueeze = False
                    if self.dataset_info.targets[target_key].is_spherical:  # spherical
                        is_invariant = (
                            int(key["o3_lambda"]) == 0 and int(key["o3_sigma"]) == 1
                        )
                        if is_invariant:
                            # squeeze components dimension
                            tensor_list = [b.values.squeeze(1) for b in block_list]
                            needs_unsqueeze = True
                        else:
                            # we don't need the targets as we will set the composition
                            # to zero
                            tensor_list = None
                    else:  # scalar
                        tensor_list = [b.values for b in block_list]

                    metadata_block = self.dataset_info.targets[target_key].layout.block(
                        key
                    )
                    if tensor_list is None:  # spherical non-invariant
                        weights_tensor = torch.zeros(
                            (
                                len(self.atomic_types),
                                len(metadata_block.components[0]),
                                len(metadata_block.properties),
                            ),
                            dtype=dtype,
                            device=device,
                        )
                    else:
                        if self.dataset_info.targets[target_key].per_atom:
                            # hack: metatensor.join doesn't work on single blocks;
                            # create TensorMaps, join, and then extract the joined block
                            joined_blocks = metatensor.torch.join(
                                [
                                    TensorMap(
                                        keys=Labels.single(),
                                        blocks=[b],
                                    )
                                    for b in block_list
                                ],
                                axis="samples",
                                remove_tensor_name=True,
                            ).block()
                            weights_tensor = metatensor.torch.sort_block(
                                metatensor.torch.mean_over_samples_block(
                                    joined_blocks,
                                    [
                                        n
                                        for n in joined_blocks.samples.names
                                        if n != "center_type"
                                    ],
                                )
                            ).values
                        else:
                            # concatenate samples, for each block
                            all_targets = torch.concatenate(tensor_list)
                            weights_tensor = _solve_linear_system(
                                composition_features, all_targets
                            )
                    if needs_unsqueeze:  # scalar invariant, needs extra dimension
                        weights_tensor = weights_tensor.unsqueeze(1)
                    weight_blocks.append(
                        TensorBlock(
                            values=weights_tensor.contiguous(),
                            # TODO: remove the .contiguous() when metatensor supports it
                            samples=Labels(
                                ["center_type"],
                                values=torch.tensor(
                                    self.atomic_types, dtype=torch.int, device=device
                                ).reshape(-1, 1),
                            ),
                            components=[
                                c.to(device) for c in metadata_block.components
                            ],
                            properties=metadata_block.properties.to(device),
                        )
                    )
                self.weights[target_key] = TensorMap(
                    keys=self.dataset_info.targets[target_key].layout.keys.to(device),
                    blocks=weight_blocks,
                )

            # make sure to update the weights buffer with the new weights
            self.register_buffer(
                target_key + "_composition_buffer",
                metatensor.torch.save_buffer(
                    self.weights[target_key].to("cpu", torch.float64)
                ).to(device),
            )

    def restart(self, dataset_info: DatasetInfo) -> "CompositionModel":
        """Restart the model with a new dataset info.

        :param dataset_info: New dataset information to be used.
        """
        for target_name, target_info in dataset_info.targets.items():
            if not self.is_valid_target(target_name, target_info):
                raise ValueError(
                    f"Composition model does not support target "
                    f"{target_name}. This is an architecture bug. "
                    "Please report this issue and help us improve!"
                )

        # merge old and new dataset info
        merged_info = self.dataset_info.union(dataset_info)
        new_atomic_types = [
            at for at in merged_info.atomic_types if at not in self.atomic_types
        ]

        if len(new_atomic_types) > 0:
            raise ValueError(
                f"New atomic types found in the dataset: {new_atomic_types}. "
                "The composition model does not support adding new atomic types."
            )

        self.new_targets = {
            target_name: target_info
            for target_name, target_info in merged_info.targets.items()
            if target_name not in self.dataset_info.targets
        }

        # register new outputs
        for target_name, target in self.new_targets.items():
            self._add_output(target_name, target)

        self.dataset_info = merged_info

        return self

    def forward(
        self,
        systems: List[System],
        outputs: Dict[str, ModelOutput],
        selected_atoms: Optional[Labels] = None,
    ) -> Dict[str, TensorMap]:
        """Compute the targets for each system based on the composition weights.

        :param systems: List of systems to calculate the energy.
        :param outputs: Dictionary containing the model outputs.
        :param selected_atoms: Optional selection of atoms for which to compute the
            predictions.
        :returns: A dictionary with the computed predictions for each system.

        :raises ValueError: If no weights have been computed or if `outputs` keys
            contain unsupported keys.
        """
        dtype = systems[0].positions.dtype
        device = systems[0].positions.device

        # move weights (TensorMaps can't be treated as buffers for now)
        self._move_weights_to_device_and_dtype(device, dtype)

        for output_name in outputs:
            if output_name not in self.weights:
                raise ValueError(
                    f"output key {output_name} is not supported by this composition "
                    "model."
                )

        # Note: atomic types are not checked. At training time, the composition model
        # is initialized with the correct types. At inference time, the checks are
        # performed by MetatensorAtomisticModel.

        # create sample labels
        sample_values_list = []
        for i_system, system in enumerate(systems):
            system_column = torch.full(
                (len(system),), i_system, dtype=torch.int, device=device
            )
            atom_column = torch.arange(len(system), device=device)
            samples_values_single_system = torch.stack(
                [system_column, atom_column], dim=1
            )
            sample_values_list.append(samples_values_single_system)
        sample_values = torch.concatenate(sample_values_list)
        sample_labels = Labels(["system", "atom"], sample_values)

        # concatenate all types for all structures
        concatenated_types = torch.concatenate([system.types for system in systems])

        # Compute the output for each system by adding the composition weights times
        # number of atoms per atomic type.
        composition_result_dict: Dict[str, TensorMap] = {}
        for output_name, output_options in outputs.items():
            blocks: List[TensorBlock] = []
            for weight_key, weight_block in self.weights[output_name].items():
                weights_tensor = self.weights[output_name].block(weight_key).values
                composition_values_per_atom = weights_tensor[
                    self.type_to_index[concatenated_types]
                ]
                blocks.append(
                    TensorBlock(
                        values=composition_values_per_atom,
                        samples=sample_labels,
                        components=weight_block.components,
                        properties=weight_block.properties,
                    )
                )
            composition_result_dict[output_name] = TensorMap(
                keys=self.weights[output_name].keys,
                blocks=blocks,
            )

            # apply selected_atoms to the composition if needed
            if selected_atoms is not None:
                composition_result_dict[output_name] = metatensor.torch.slice(
                    composition_result_dict[output_name], "samples", selected_atoms
                )

            if not output_options.per_atom:  # sum over atoms if needed
                composition_result_dict[output_name] = sum_over_atoms(
                    composition_result_dict[output_name]
                )

        return composition_result_dict

    def _add_output(self, target_name: str, target_info: TargetInfo) -> None:
        self.outputs[target_name] = ModelOutput(
            quantity=target_info.quantity,
            unit=target_info.unit,
            per_atom=True,
        )
        self.weights[target_name] = TensorMap(
            keys=target_info.layout.keys,
            blocks=[
                TensorBlock(
                    values=torch.zeros(
                        ([len(self.atomic_types)] + b.shape[1:]),
                        dtype=torch.float64,
                    ),
                    samples=Labels(
                        names=["center_type"],
                        values=torch.tensor(self.atomic_types, dtype=torch.int).reshape(
                            -1, 1
                        ),
                    ),
                    components=b.components,
                    properties=b.properties,
                )
                for b in target_info.layout.blocks()
            ],
        )

        # register a buffer to store the weights; this is necessary because the weights
        # are TensorMaps and cannot be stored in the state_dict
        fake_weights = TensorMap(
            keys=self.dataset_info.targets[target_name].layout.keys,
            blocks=[
                TensorBlock(
                    values=torch.zeros(
                        (len(self.atomic_types),) + b.values.shape[1:],
                        dtype=torch.float64,
                    ),
                    samples=Labels(
                        names=["center_type"],
                        values=torch.tensor(self.atomic_types, dtype=torch.int).reshape(
                            -1, 1
                        ),
                    ),
                    components=b.components,
                    properties=b.properties,
                )
                for b in target_info.layout.blocks()
            ],
        )
        self.register_buffer(
            target_name + "_composition_buffer",
            metatensor.torch.save_buffer(fake_weights),
        )

    def _move_weights_to_device_and_dtype(
        self, device: torch.device, dtype: torch.dtype
    ):
        if len(self.weights) != 0:
            if self.weights[list(self.weights.keys())[0]].device != device:
                self.weights = {k: v.to(device) for k, v in self.weights.items()}
            if self.weights[list(self.weights.keys())[0]].dtype != dtype:
                self.weights = {k: v.to(dtype) for k, v in self.weights.items()}

    @staticmethod
    def is_valid_target(target_name: str, target_info: TargetInfo) -> bool:
        """Finds if a ``TargetInfo`` object is compatible with a composition model.

        :param target_info: The ``TargetInfo`` object to be checked.
        """
        # only scalars can have composition contributions
        if not target_info.is_scalar and not target_info.is_spherical:
            logging.debug(
                f"Composition model does not support target {target_name} "
                "since it is not either scalar or spherical."
            )
            return False
        if (
            target_info.is_spherical
            and len(target_info.layout.blocks({"o3_lambda": 0, "o3_sigma": 1})) == 0
        ):
            logging.debug(
                f"Composition model does not support spherical target {target_name} "
                "since it does not have any invariant blocks."
            )
            return False
        return True

    def sync_tensor_maps(self):
        # Reload the weights of the (old) targets, which are not stored in the model
        # state_dict, from the buffers
        for k in self.dataset_info.targets:
            self.weights[k] = metatensor.torch.load_buffer(
                self.__getattr__(k + "_composition_buffer")
            )


def _solve_linear_system(composition_features, all_targets) -> torch.Tensor:
    compf_t_at_compf = composition_features.T @ composition_features
    compf_t_at_targets = composition_features.T @ all_targets
    trace_magnitude = float(torch.diag(compf_t_at_compf).abs().mean())
    regularizer = 1e-14 * trace_magnitude
    return torch.linalg.solve(
        compf_t_at_compf
        + regularizer
        * torch.eye(
            composition_features.shape[1],
            dtype=composition_features.dtype,
            device=composition_features.device,
        ),
        compf_t_at_targets,
    )



from typing import Dict, List, Union

import metatensor.torch
import numpy as np
import torch
from metatensor.torch import TensorMap
from metatensor.torch.atomistic import ModelOutput

from .additive import remove_additive
from .data import Dataset, DatasetInfo, TargetInfo, get_all_targets
from .jsonschema import validate
from .per_atom import average_by_num_atoms
from .transfer import systems_and_targets_to_device


class Scaler(torch.nn.Module):
    """
    A class that scales the targets of regression problems to unit standard
    deviation.

    In most cases, this should be used in conjunction with a composition model
    (that removes the multi-dimensional "mean" across the composition space) and/or
    other additive models. See the `train_model` method for more details.

    The scaling is performed per-atom, i.e., in cases where the targets are
    per-structure, the standard deviation is calculated on the targets divided by
    the number of atoms in each structure.

    :param model_hypers: A dictionary of model hyperparameters. The paramater is ignored
        and is only present to be consistent with the general model API.
    :param dataset_info: An object containing information about the dataset, including
        target quantities and atomic types.
    """

    outputs: Dict[str, ModelOutput]
    scales: torch.Tensor

    def __init__(self, model_hypers: Dict, dataset_info: DatasetInfo):
        super().__init__()

        # `model_hypers` should be an empty dictionary
        validate(
            instance=model_hypers,
            schema={"type": "object", "additionalProperties": False},
        )

        self.dataset_info = dataset_info

        self.new_targets: Dict[str, TargetInfo] = dataset_info.targets
        self.outputs: Dict[str, ModelOutput] = {}

        # Initially, the scales are empty. They will be expanded as new outputs
        # are registered with `_add_output`.
        self.register_buffer("scales", torch.ones((0,), dtype=torch.float64))
        self.output_name_to_output_index: Dict[str, int] = {}
        for target_name, target_info in self.dataset_info.targets.items():
            self._add_output(target_name, target_info)

    def train_model(
        self,
        datasets: List[Union[Dataset, torch.utils.data.Subset]],
        additive_models: List[torch.nn.Module],
        treat_as_additive: bool,
    ) -> None:
        """
        Calculate the scaling weights for all the targets in the datasets.

        :param datasets: Dataset(s) to calculate the scaling weights for.
        :param additive_models: Additive models to be removed from the targets
            before calculating the statistics.
        :param treat_as_additive: If True, all per-structure targets (i.e. those that)
            do not contain an ``atom`` label name, are treated as additive.

        :raises ValueError: If the provided datasets contain targets unknown
            to the scaler or if the targets are not treated as additive.
        """
        if not treat_as_additive:
            raise ValueError(
                "The Scaler class can currently only be trained by treating targets "
                "as additive."
            )

        if not isinstance(datasets, list):
            datasets = [datasets]

        device = self.scales.device

        # Fill the scales for each "new" target (i.e. those that do not already
        # have scales from a previous training run)
        for target_key in self.new_targets:
            datasets_with_target = []
            for dataset in datasets:
                if target_key in get_all_targets(dataset):
                    datasets_with_target.append(dataset)
            if len(datasets_with_target) == 0:
                raise ValueError(
                    f"Target {target_key} in the model's new capabilities is not "
                    "present in any of the training datasets."
                )

            sum_of_squared_targets = 0.0
            total_num_elements = 0
            for dataset in datasets_with_target:
                for sample in dataset:
                    systems = [sample["system"]]
                    targets = {target_key: sample[target_key]}

                    systems, targets = systems_and_targets_to_device(
                        systems, targets, device
                    )

                    for additive_model in additive_models:
                        target_info_dict = {target_key: self.new_targets[target_key]}
                        targets = remove_additive(
                            systems,
                            targets,
                            additive_model,
                            target_info_dict,
                        )

                    # calculate standard deviations on per-atom quantities
                    targets = average_by_num_atoms(
                        targets,
                        systems,
                        per_structure_keys=[],
                    )

                    target_info = self.new_targets[target_key]
                    if (
                        target_info.quantity == "energy"
                        and "positions" in target_info.gradients
                    ):
                        # special case: here we want to scale with respect to the forces
                        # rather than the energies
                        sum_of_squared_targets += torch.sum(
                            targets[target_key].block().gradient("positions").values
                            ** 2
                        ).item()
                        total_num_elements += (
                            targets[target_key]
                            .block()
                            .gradient("positions")
                            .values.numel()
                        )
                    else:
                        sum_of_squared_targets += sum(
                            torch.sum(block.values**2).item()
                            for block in targets[target_key].blocks()
                        )
                        total_num_elements += sum(
                            block.values.numel()
                            for block in targets[target_key].blocks()
                        )

            self.scales[self.output_name_to_output_index[target_key]] = np.sqrt(
                sum_of_squared_targets / total_num_elements
            )

    def restart(self, dataset_info: DatasetInfo) -> "Scaler":
        # merge old and new dataset info
        merged_info = self.dataset_info.union(dataset_info)

        self.new_targets = {
            key: value
            for key, value in merged_info.targets.items()
            if key not in self.dataset_info.targets
        }

        # register new outputs
        for target_name, target in self.new_targets.items():
            self._add_output(target_name, target)

        self.dataset_info = merged_info

        return self

    def forward(
        self,
        outputs: Dict[str, TensorMap],
    ) -> Dict[str, TensorMap]:
        """
        Scales all the targets in the outputs dictionary back to their
        original scale.

        :param outputs: A dictionary of target quantities and their values
            to be scaled.

        :raises ValueError: If an output does not have a corresponding
            scale in the scaler model.
        """
        scaled_outputs: Dict[str, TensorMap] = {}
        for target_key, target in outputs.items():
            if target_key in self.outputs:
                scale = float(
                    self.scales[self.output_name_to_output_index[target_key]].item()
                )
                scaled_target = metatensor.torch.multiply(target, scale)
                scaled_outputs[target_key] = scaled_target
            else:
                scaled_outputs[target_key] = target

        return scaled_outputs

    def _add_output(self, target_name: str, target_info: TargetInfo) -> None:
        self.outputs[target_name] = ModelOutput(
            quantity=target_info.quantity,
            unit=target_info.unit,
            per_atom=True,
        )

        self.scales = torch.cat(
            [self.scales, torch.tensor([1.0], dtype=self.scales.dtype)]
        )
        self.output_name_to_output_index[target_name] = len(self.scales) - 1

    def get_scales_dict(self) -> Dict[str, float]:
        """
        Return a dictionary with the scales for each output and output gradient.

        :return: A dictionary with the scales for each output and output gradient.
            These correspond to the standard deviation of the targets in the
            original dataset. The scales for each output gradient are the same
            as the corresponding output.
        """

        scales_dict = {
            output_name: self.scales[output_index].item()
            for output_name, output_index in self.output_name_to_output_index.items()
        }
        # Add gradients if present. They have the same scale as the corresponding output
        for output_name in list(scales_dict.keys()):
            gradient_names_for_output = self.dataset_info.targets[output_name].gradients
            for gradient_name in gradient_names_for_output:
                scales_dict[output_name + "_" + gradient_name + "_gradients"] = (
                    scales_dict[output_name]
                )
        return scales_dict


def remove_scale(
    targets: Dict[str, TensorMap],
    scaler: Scaler,
):
    """
    Scale all targets to a standard deviation of one.

    :param targets: Dictionary containing the targets to be scaled.
    :param scaler: The scaler used to scale the targets.
    """
    scaled_targets = {}
    for target_key in targets.keys():
        scale = float(
            scaler.scales[scaler.output_name_to_output_index[target_key]].item()
        )
        scaled_targets[target_key] = metatensor.torch.multiply(
            targets[target_key], 1.0 / scale
        )

    return scaled_targets



from typing import Dict


def update_state_dict(state_dict: Dict) -> Dict:
    """
    Updates the state_dict keys so they match the model's keys.
    """
    new_state_dict = {}
    for name, value in state_dict.items():
        if "pet_model." in name:
            name = name.split("pet_model.")[1]
        new_state_dict[name] = value
    return new_state_dict



from typing import List

import torch
from metatensor.torch.atomistic import System


class LongRangeFeaturizer(torch.nn.Module):
    """A class to compute long-range features starting from short-range features.

    :param hypers: Dictionary containing the hyperparameters for the long-range
        featurizer.
    :param feature_dim: The dimension of the short-range features (which also
        corresponds to the number of long-range features that will be returned).
    :param neighbor_list_options: A :py:class:`NeighborListOptions` object containing
        the neighbor list information for the short-range model.
    """

    def __init__(self, hypers, feature_dim, neighbor_list_options):
        super(LongRangeFeaturizer, self).__init__()

        try:
            from torchpme import CoulombPotential
            from torchpme.calculators import Calculator, EwaldCalculator, P3MCalculator
        except ImportError:
            raise ImportError(
                "`torch-pme` is required for long-range models. "
                "Please install it with `pip install torch-pme`."
            )

        self.ewald_calculator = EwaldCalculator(
            potential=CoulombPotential(
                smearing=hypers["smearing"],
                exclusion_radius=neighbor_list_options.cutoff,
            ),
            full_neighbor_list=neighbor_list_options.full_list,
            lr_wavelength=hypers["kspace_resolution"],
        )
        self.p3m_calculator = P3MCalculator(
            potential=CoulombPotential(
                smearing=hypers["smearing"],
                exclusion_radius=neighbor_list_options.cutoff,
            ),
            interpolation_nodes=hypers["interpolation_nodes"],
            full_neighbor_list=neighbor_list_options.full_list,
            mesh_spacing=hypers["kspace_resolution"],
        )
        self.use_ewald = hypers["use_ewald"]
        self.direct_calculator = Calculator(
            potential=CoulombPotential(
                smearing=None,
                exclusion_radius=neighbor_list_options.cutoff,
            ),
            full_neighbor_list=False,  # see docs of torch.combinations
        )

        self.neighbor_list_options = neighbor_list_options
        self.charges_map = torch.nn.Linear(feature_dim, feature_dim)

    def forward(
        self,
        systems: List[System],
        features: torch.Tensor,
        neighbor_distances: torch.Tensor,
    ) -> torch.Tensor:
        """Compute the long-range features for a list of systems.

        :param systems: A list of :py:class:`System` objects for which to compute the
            long-range features. Each system must contain a neighbor list consistent
            with the neighbor list options used to create the class.
        :param features: A tensor of short-range features for the systems.
        :param neighbor_distances: A tensor of neighbor distances for the systems,
            which must be consistent with the neighbor list options used to create the
            class.
        """
        charges = self.charges_map(features)

        last_len_nodes = 0
        last_len_edges = 0
        long_range_features = []
        for system in systems:
            system_charges = charges[last_len_nodes : last_len_nodes + len(system)]
            last_len_nodes += len(system)

            neighbor_list = system.get_neighbor_list(self.neighbor_list_options)
            neighbor_indices_system = neighbor_list.samples.view(
                ["first_atom", "second_atom"]
            ).values

            neighbor_distances_system = neighbor_distances[
                last_len_edges : last_len_edges + len(neighbor_indices_system)
            ]
            last_len_edges += len(neighbor_indices_system)

            if system.pbc.any() and not system.pbc.all():
                raise NotImplementedError(
                    "Long-range features are not currently supported for systems "
                    "with mixed periodic and non-periodic boundary conditions."
                )

            if system.pbc.all():  # periodic
                if self.use_ewald and self.training:  # use Ewald for training only
                    potential = self.ewald_calculator.forward(
                        charges=system_charges,
                        cell=system.cell,
                        positions=system.positions,
                        neighbor_indices=neighbor_indices_system,
                        neighbor_distances=neighbor_distances_system,
                    )
                else:
                    potential = self.p3m_calculator.forward(
                        charges=system_charges,
                        cell=system.cell,
                        positions=system.positions,
                        neighbor_indices=neighbor_indices_system,
                        neighbor_distances=neighbor_distances_system,
                    )
            else:  # non-periodic
                # compute the distance between all pairs of atoms
                neighbor_indices_system = torch.combinations(
                    torch.arange(len(system), device=system.positions.device), 2
                )
                neighbor_distances_system = torch.sqrt(
                    torch.sum(
                        (
                            system.positions[neighbor_indices_system[:, 1]]
                            - system.positions[neighbor_indices_system[:, 0]]
                        )
                        ** 2,
                        dim=1,
                    )
                )
                potential = self.direct_calculator.forward(
                    charges=system_charges,
                    cell=system.cell,
                    positions=system.positions,
                    neighbor_indices=neighbor_indices_system,
                    neighbor_distances=neighbor_distances_system,
                )
            long_range_features.append(potential * system_charges)
        return torch.concatenate(long_range_features)


class DummyLongRangeFeaturizer(torch.nn.Module):
    # a dummy class for torchscript
    def __init__(self):
        super().__init__()
        self.use_ewald = True

    def forward(
        self,
        systems: List[System],
        features: torch.Tensor,
        neighbor_distances: torch.Tensor,
    ) -> torch.Tensor:
        return torch.tensor(0)
